#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Shape class for p5.
"""

import contextlib
import functools

import numpy as np
import math

from p5.core.color import Color
from p5.core.constants import SType
from p5.pmath import matrix
from p5.pmath.vector import Point
from p5.pmath.utils import SINCOS
from p5.core import p5

__all__ = ["PShape"]


def _ensure_editable(func):
    """A decorater that ensures that a shape is in 'edit' mode."""

    @functools.wraps(func)
    def editable_method(instance, *args, **kwargs):
        if not instance._in_edit_mode:
            raise ValueError("{} only works in edit mode".format(func.__name__))
        return func(instance, *args, **kwargs)

    return editable_method


def _apply_transform(func):
    """Apply the matrix transformation to the shape."""

    @functools.wraps(func)
    def mfunc(instance, *args, **kwargs):
        tmat = func(instance, *args, **kwargs)
        instance._matrix = instance._matrix.dot(tmat)
        return tmat

    return mfunc


def _call_on_children(func):
    """Call the method on all child shapes"""

    @functools.wraps(func)
    def rfunc(instance, *args, **kwargs):
        rval = func(instance, *args, **kwargs)
        for child in instance.children:
            rfunc(child, *args, **kwargs)
        return rval

    return rfunc


class PShape:
    """Custom shape class for p5.

    :param vertices: List of (polygonal) vertices for the shape.
    :type vertices: list | np.ndarray

    :param fill_color: Fill color of the shape (default: 'auto' i.e.,
        the current renderer fill)
    :type fill_color: 'auto' | None | tuple | p5.Color

    :param stroke_color: Stroke color of the shape (default: 'auto'
        i.e., the current renderer stroke color)
    :type stroke_color: 'auto' | None | tuple | p5.color

    :param visible: toggles shape visibility (default: False)
    :type visible: bool

    :param children: List of sub-shapes for the current shape
        (default: [])
    :type children: list

    """

    def __init__(
        self,
        fill_color="auto",
        stroke_color="auto",
        stroke_weight="auto",
        stroke_join="auto",
        stroke_cap="auto",
        visible=False,
        children=None,
        contours=tuple(),
        vertices=tuple(),
        shape_type=SType.TESS,
    ):
        # basic properties of the shape
        self._fill = None
        self._stroke = None
        self._stroke_weight = None
        self._stroke_cap = None
        self._stroke_join = None

        self._matrix = np.identity(4)
        self._transform_matrix = np.identity(4)
        self._transformed_draw_vertices = None

        # a flag to check if the shape is being edited right now.
        self._in_edit_mode = False

        self.fill = fill_color
        self.stroke = stroke_color
        self.stroke_weight = stroke_weight
        self.stroke_cap = stroke_cap
        self.stroke_join = stroke_join

        self.children = children or []
        self.visible = visible

        self.vertices = list(vertices)
        self.shape_type = shape_type
        self.contours = [list(c) for c in contours]  # List of all contours

    def _set_color(self, name, value=None):
        color = None

        if isinstance(value, Color):  # Is Color, no need to parse
            color = value
        else:  # Not Color, attempt to parse it
            if name == "stroke" and p5.renderer.style.stroke_enabled:
                color = Color(
                    *p5.renderer.style.stroke_color, color_mode="RGBA", normed=True
                )
            if name == "fill" and p5.renderer.style.fill_enabled:
                color = Color(
                    *p5.renderer.style.fill_color, color_mode="RGBA", normed=True
                )

        if name == "stroke":
            self._stroke = color
        elif name == "fill":
            self._fill = color

    @property
    def fill(self):
        return self._fill

    @fill.setter
    def fill(self, new_color):
        self._set_color("fill", new_color)

    @property
    def stroke(self):
        return self._stroke

    @stroke.setter
    def stroke(self, new_color):
        self._set_color("stroke", new_color)

    @property
    def stroke_weight(self):
        return self._stroke_weight

    @stroke_weight.setter
    def stroke_weight(self, stroke):
        if stroke == "auto":
            self._stroke_weight = p5.renderer.style.stroke_weight
        else:
            self._stroke_weight = stroke

    @property
    def stroke_join(self):
        return self._stroke_join

    @stroke_join.setter
    def stroke_join(self, stroke):
        if stroke == "auto":
            self._stroke_join = p5.renderer.style.stroke_join
        else:
            self._stroke_join = stroke

    @property
    def stroke_cap(self):
        return self._stroke_cap

    @stroke_cap.setter
    def stroke_cap(self, stroke):
        if stroke == "auto":
            self._stroke_cap = p5.renderer.style.stroke_cap
        else:
            self._stroke_cap = stroke

    @contextlib.contextmanager
    def edit(self, reset=True):
        """Put the shape in edit mode.

        :param reset: Toggles whether the shape should be "reset"
            during editing. When set to `True` all existing shape
            vertices are cleared. When set to `False` the new vertices
            are appended at the end of the existing vertex list.
            (default: True)
        :type reset: bool

        :raises ValueError: if the shape is already being edited.

        """
        if self._in_edit_mode:
            raise ValueError("Shape is being edited already")
        if reset:
            self.vertices = []
            self.contours = []
        self._in_edit_mode = True
        yield
        self._in_edit_mode = False

    @_ensure_editable
    def add_vertex(self, vertex):
        """Add a vertex to the current shape

        :param vertex: The (next) vertex to add to the current shape.
        :type vertex: tuple | list | p5.Vector | np.ndarray
        """
        self.vertices.append(Point(*vertex))

    @_ensure_editable
    def update_vertex(self, idx, vertex):
        """Edit an individual vertex.

        :param idx: index of the vertex to be edited
        :type idx: int

        :param vertex: The (next) vertex to add to the current shape.
        :type vertex: tuple | list | p5.Vector | np.ndarray
        """
        self.vertices[idx] = Point(*vertex)

    def add_child(self, child):
        """Add a child shape to the current shape

        :param child: Child to be added
        :type child: PShape
        """
        self.children.append(child)

    def transform_matrix(self, mat):
        self._transform_matrix = mat

    @property
    def child_count(self):
        """Number of children.

        :returns: The current number of children.
        :rtype: int
        """
        return len(self.children)

    def apply_matrix(self, mat):
        """Apply the given transformation matrix to the shape.

        :param mat: the 4x4 matrix to be applied to the current shape.
        :type mat: (4, 4) np.ndarray

        """
        self._matrix = self._matrix.dot(mat)

    @_call_on_children
    def apply_transform_matrix(self, mat):
        self._matrix = self._matrix.dot(mat)

    @_call_on_children
    def reset_matrix(self):
        """Reset the transformation matrix associated with the shape."""
        self._matrix = np.identity(4)

    @_call_on_children
    @_apply_transform
    def translate(self, x, y, z=0):
        """Translate the shape origin to the given location.

        :param x: The displacement amount in the x-direction (controls
            the left/right displacement)
        :type x: int

        :param y: The displacement amount in the y-direction (controls
            the up/down displacement)
        :type y: int

        :param z: The displacement amount in the z-direction (0 by
            default). This controls the displacement away-from/towards
            the screen.
        :type z: int

        :returns: The translation matrix applied to the transform
            matrix.
        :rtype: np.ndarray

        """
        tmat = matrix.translation_matrix(x, y, z)
        return tmat

    @_call_on_children
    @_apply_transform
    def rotate(self, theta, axis=(0, 0, 1)):
        """Rotate the shape by the given angle along the given axis.

        :param theta: The angle by which to rotate (in radians)
        :type theta: float

        :param axis: The axis along which to rotate (defaults to the
            z-axis)
        :type axis: np.ndarray | list

        :returns: The rotation matrix used to apply the
            transformation.
        :rtype: np.ndarray

        """
        axis = np.array(axis[:])
        tmat = matrix.rotation_matrix(axis, theta)
        return tmat

    def rotate_x(self, theta):
        """Rotate the shape along the x axis.

        :param theta: angle by which to rotate (in radians)
        :type theta: float

        :returns: The rotation matrix used to apply the
            transformation.
        :rtype: np.ndarray

        """
        return self.rotate(theta, axis=[1, 0, 0])

    def rotate_y(self, theta):
        """Rotate the shape along the y axis.

        :param theta: angle by which to rotate (in radians)
        :type theta: float

        :returns: The rotation matrix used to apply the
             transformation.
        :rtype: np.ndarray

        """
        return self.rotate(theta, axis=[0, 1, 0])

    def rotate_z(self, theta):
        """Rotate the shape along the z axis.

        :param theta: angle by which to rotate (in radians)
        :type theta: float

        :returns: The rotation matrix used to apply the
            transformation.
        :rtype: np.ndarray

        """
        return self.rotate(theta)

    @_call_on_children
    @_apply_transform
    def scale(self, sx, sy=None, sz=None):
        """Scale the shape by the given factor.

        :param sx: scale factor along the x-axis.
        :type sx: float

        :param sy: scale factor along the y-axis (defaults to None)
        :type sy: float

        :param sz: scale factor along the z-axis (defaults to None)
        :type sz: float

        :returns: The transformation matrix used to appy the
            transformation.
        :rtype: np.ndarray

        """
        if sy is None and sz is None:
            sy = sx
            sz = sx
        elif sz is None:
            sz = 1
        tmat = matrix.scale_transform(sx, sy, sz)
        return tmat

    @_call_on_children
    @_apply_transform
    def shear_x(self, theta):
        """Shear shape along the x-axis.

        :param theta: angle to shear by (in radians)
        :type theta: float

        :returns: The shear matrix used to apply the tranformation.
        :rtype: np.ndarray

        """
        shear_mat = np.identity(4)
        shear_mat[0, 1] = np.tan(theta)
        return shear_mat

    @_call_on_children
    @_apply_transform
    def shear_y(self, theta):
        """Shear shape along the y-axis.

        :param theta: angle to shear by (in radians)
        :type theta: float

        :returns: The shear matrix used to apply the transformation.
        :rtype: np.ndarray

        """
        shear_mat = np.identity(4)
        shear_mat[1, 0] = np.tan(theta)
        return shear_mat


# We use these in ellipse tessellation. The algorithm is similar to
# the one used in Processing and the we compute the number of
# subdivisions per ellipse using the following formula:
#
#    min(M, max(N, (2 * pi * size / F)))
#
# Where,
#
# - size :: is the measure of the dimensions of the circle when
#   projected in screen coordiantes.
#
# - F :: sets the minimum number of subdivisions. A smaller `F` would
#   produce more detailed circles (== POINT_ACCURACY_FACTOR)
#
# - N :: Minimum point accuracy (== MIN_POINT_ACCURACY)
#
# - M :: Maximum point accuracy (== MAX_POINT_ACCURACY)
#
MIN_POINT_ACCURACY = 20
MAX_POINT_ACCURACY = 200
POINT_ACCURACY_FACTOR = 10


class Arc(PShape):
    def __init__(
        self,
        center,
        radii,
        start_angle,
        stop_angle,
        mode=None,
        fill_color="auto",
        stroke_color="auto",
        stroke_weight="auto",
        stroke_join="auto",
        stroke_cap="auto",
        **kwargs
    ):
        self._center = center
        self._radii = radii
        self._start_angle = start_angle
        self._stop_angle = stop_angle
        self.arc_mode = mode

        gl_type = SType.TESS if mode in ["OPEN", "CHORD"] else SType.TRIANGLE_FAN
        super().__init__(
            fill_color=fill_color,
            stroke_color=stroke_color,
            stroke_weight=stroke_weight,
            stroke_join=stroke_join,
            stroke_cap=stroke_cap,
            shape_type=gl_type,
            **kwargs
        )
        self._tessellate()

    def _tessellate(self):
        """Generate vertex and face data using radii."""
        rx = self._radii[0]
        ry = self._radii[1]

        c1x = self._center[0]
        c1y = self._center[1]
        s1 = p5.renderer.transform_matrix.dot(np.array([c1x, c1y, 0, 1]))

        c2x = c1x + rx
        c2y = c1y + ry
        s2 = p5.renderer.transform_matrix.dot(np.array([c2x, c2y, 0, 1]))

        sdiff = s2 - s1
        size_acc = (
            np.sqrt(np.sum(sdiff * sdiff)) * math.pi * 2
        ) / POINT_ACCURACY_FACTOR

        acc = min(MAX_POINT_ACCURACY, max(MIN_POINT_ACCURACY, int(size_acc)))
        inc = int(len(SINCOS) / acc)

        sclen = len(SINCOS)
        start_index = int((self._start_angle / (math.pi * 2)) * sclen)
        end_index = int((self._stop_angle / (math.pi * 2)) * sclen)

        vertices = [(c1x, c1y, 0)] if self.arc_mode in ["PIE", None] else []
        for idx in range(start_index, end_index, inc):
            i = idx % sclen
            vertices.append((c1x + rx * SINCOS[i][1], c1y + ry * SINCOS[i][0], 0))
        vertices.append(
            (
                c1x + rx * SINCOS[end_index % sclen][1],
                c1y + ry * SINCOS[end_index % sclen][0],
                0,
            )
        )
        if self.arc_mode == "CHORD" or self.arc_mode == "PIE":
            vertices.append(vertices[0])
        self.vertices = vertices



"""
Part of the io library. Includes functions that require **internet** connection.
"""

from urllib import request
import json
import requests

# Synchronous


class http_get:
    """
    Performs a synchronous GET request on the given url.

    :param url: url of the API
    :type url: string

    """

    def __init__(self, url):
        resp = request.urlopen(url)

        data = resp.read().decode("UTF-8")

        self.text = data

        try:
            self.json = json.loads(data)
        except json.JSONDecodeError:
            self.json = data


# Synchronous


def http_post(url, post_data):
    """
    Performs a synchronous POST request on the given url.

    :param url: url of the API
    :type url: string
    :param post_data: Key-value pairs of the request payload.
    :type post_data: dict

    """
    data = requests.post(url, post_data)
    data = data.json()
    return data



# This file adds Processing API compatibility

from .curves import (
    bezier_detail,
    bezier_point,
    bezier_tangent,
    curve_point,
    curve_tightness,
    curve_detail,
    curve_tangent,
    quadratic_point,
)
from .rand import noise_detail, noise_seed, random_uniform, random_seed, random_gaussian


def bezierDetail(detailValue):
    """Change the resolution used to draw bezier curves.

    :param detailValue: New resolution to be used.
    :type detailValue: int
    """
    bezier_detail(detailValue)


def bezierPoint(start, control1, control2, stop, parameter):
    """Return the coordinate of a point along a bezier curve.

    :param start: The start point of the bezier curve
    :type start: float or n-tuple

    :param control1: The first control point of the bezier curve
    :type control1: float or n-tuple

    :param control2: The second control point of the bezier curve
    :type control2: float or n-tuple

    :param stop: The end point of the bezier curve
    :type stop: float or n-tuple

    :param parameter: The parameter for the required location along
        the curve. Should be in the range [0.0, 1.0] where 0 indicates
        the start of the curve and 1 indicates the end of the curve.
    :type parameter: float

    :returns: The coordinate of the point along the bezier curve.
    :rtype: float or n-tuple

    """
    return bezier_point(start, control1, control2, stop, parameter)


def bezierTangent(start, control1, control2, stop, parameter):
    """Return the tangent at a point along a bezier curve.

    :param start: The start point of the bezier curve
    :type start: float or n-tuple.

    :param control1: The first control point of the bezier curve
    :type control1: float or n-tuple.

    :param control2: The second control point of the bezier curve
    :type control2: float or n-tuple.

    :param stop: The end point of the bezier curve
    :type stop: float or n-tuple.

    :param parameter: The parameter for the required tangent location
        along the curve. Should be in the range [0.0, 1.0] where 0
        indicates the start of the curve and 1 indicates the end of
        the curve.
    :type parameter: float

    :returns: The tangent at the required point along the bezier
        curve.
    :rtype: float or n-tuple

    """
    return bezier_tangent(start, control1, control2, stop, parameter)


def curveDetail(detailValue):
    """Change the resolution used to draw bezier curves.

    :param detailValue: New resolution to be used.
    :type detailValue: int
    """
    curve_detail(detailValue)


def curveTightness(amount):
    """Change the curve tightness used to draw curves.

    :param amount: new curve tightness amount.
    :type amount: int
    """
    curve_tightness(amount)


def curvePoint(point1, point2, point3, point4, parameter):
    """Return the coordinates of a point along a curve.

    :param point1: The first control point of the curve.
    :type point1: float or n-tuple.

    :param point2: The second control point of the curve.
    :type point2: float or n-tuple.

    :param point3: The third control point of the curve.
    :type point3: float or n-tuple.

    :param point4: The fourth control point of the curve.
    :type point4: float or n-tuple.

    :param parameter: The parameter for the required point location
        along the curve. Should be in the range [0.0, 1.0] where 0
        indicates the start of the curve and 1 indicates the end of
        the curve.
    :type parameter: float

    :returns: The coordinate of the point at the required location
        along the curve.
    :rtype: float or n-tuple

    """
    return curve_point(point1, point2, point3, point4, parameter)


def curveTangent(point1, point2, point3, point4, parameter):
    """Return the tangent at a point along a curve.

    :param point1: The first control point of the curve.
    :type point1: float or n-tuple.

    :param point2: The second control point of the curve.
    :type point2: float or n-tuple.

    :param point3: The third control point of the curve.
    :type point3: float or n-tuple.

    :param point4: The fourth control point of the curve.
    :type point4: float or n-tuple.

    :param parameter: The parameter for the required tangent location
        along the curve. Should be in the range [0.0, 1.0] where 0
        indicates the start of the curve and 1 indicates the end of
        the curve.
    :type parameter: float

    :returns: The tangent at the required point along the curve.
    :rtype: float or n-tuple

    """
    return curve_tangent(point1, point2, point3, point4, parameter)


def quadraticPoint(start, control, stop, parameter):
    """Return the coordinates of a point along a bezier curve.

    :param point1: The start point of the curve.
    :type point1: float or n-tuple.

    :param point3: The control point of the curve.
    :type point3: float or n-tuple.

    :param point4: The end point of the curve.
    :type point4: float or n-tuple.

    :param parameter: The parameter for the required point location
        along the curve. Should be in the range [0.0, 1.0] where 0
        indicates the start of the curve and 1 indicates the end of
        the curve.
    :type parameter: float

    :returns: The coordinate of the point at the required location
        along the curve.
    :rtype: float or n-tuple

    """
    return quadratic_point(start, control, stop, parameter)


def noiseDetail(octaves=4, falloff=0.5):
    """Adjust the level of noise detail produced by noise().

    :param octaves: The number of octaves to compute the noise for
        (defaults to 4).
    :type octaves: int

    :param falloff:
    :type falloff: float

    :note: For :code:`falloff` values greater than 0.5,
        :code:`noise()` will return values greater than 1.0.

    """
    noise_detail(octaves, falloff)


def noiseSeed(seed):
    """Set the seed value for :code:`noise()`

    By default :code:`noise()` produes different values each time the
    sketch is run. Setting the :code:`seed` parameter to a constant
    will make :code:`noise()` return the same values each time the
    sketch is run.

    :param seed: The required seed value.
    :type seed: int

    """
    noise_seed(seed)


def randomUniform(high=1, low=0):
    """Return a uniformly sampled random number.

    :param high: The upper limit on the random value (defaults to 1).
    :type high: float

    :param low: The lowe limit on the random value (defaults to 0).
    :type low: float

    :returns: A random number between :code:`low` and :code:`high`.
    :rtype: float

    """
    return random_uniform(high, low)


def randomGaussian(mean=0, stdDev=1):
    """Return a normally sampled random number.

    :param mean: The mean value to be used for the normal distribution
        (defaults to 0).
    :type mean: float

    :param std_dev: The standard deviation to be used for the normal
        distribution (defaults to 1).
    :type std_dev: float

    :returns: A random number selected from a normal distribution with
        the given :code:`mean` and :code:`std_dev`.
    :rtype: float

    """
    return random_gaussian(mean, stdDev)


def randomSeed(seed):
    """Set the seed used to generate random numbers.

    :param seed: The required seed value.
    :type seed: int
    """
    random_seed(seed)



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import builtins

from p5.pmath.vector import Point
from . import p5
from .constants import TESS
from ..pmath import curves
import copy

shape_kind = None
vertices = []  # stores the vertex coordinates
vertices_types = []  # stores the type of vertex. Eg: bezier, curve, etc
curr_contour_vertices = []
curr_contour_vertices_types = []
contour_vertices = []  # list of all contours [[v1, v2, ...], [v1, v2, ...]]
contour_vertices_types = []  # list of all vertex types [[t1, t2, ...], [t1, t2, ...]]
is_bezier = False
is_curve = False
is_quadratic = False
is_contour = False
in_contour = False
is_first_contour = True

__all__ = [
    "begin_shape",
    "end_shape",
    "begin_contour",
    "end_contour",
    "curve_vertex",
    "bezier_vertex",
    "quadratic_vertex",
    "vertex",
]


def begin_shape(kind=TESS):
    """Begin shape drawing.  This is a helpful way of generating custom shapes quickly.

    :param kind: TESS, POINTS, LINES, TRIANGLES, TRIANGLE_FAN, TRIANGLE_STRIP, QUADS, or QUAD_STRIP; defaults to TESS
    :type kind: SType
    """
    global shape_kind, vertices, contour_vertices, vertices_types, contour_vertices_types, is_contour
    global curr_contour_vertices, curr_contour_vertices_types

    shape_kind = kind
    is_contour = False
    vertices = []
    vertices_types = []
    contour_vertices = []
    contour_vertices_types = []
    curr_contour_vertices = []
    curr_contour_vertices_types = []


def curve_vertex(x, y, z=0):
    """
    Specifies vertex coordinates for curves. The first
    and last points in a series of curveVertex() lines
    will be used to guide the beginning and end of a the
    curve. A minimum of four points is required to draw a
    tiny curve between the second and third points. Adding
    a fifth point with curveVertex() will draw the curve
    between the second, third, and fourth points. The
    curveVertex() function is an implementation of
    Catmull-Rom splines.

    :param x: x-coordinate of the vertex
    :type x: float

    :param y: y-coordinate of the vertex
    :type y: float

    :param z: z-coordinate of the vertex
    :type z: float
    """

    global is_curve
    is_curve = True

    if p5.mode == "3D":
        return
    if builtins.current_renderer == "vispy":
        if is_contour:
            curr_contour_vertices.append((x, y, z))
            curr_contour_vertices_types.append(2)
        else:
            vertices.append((x, y, z))  # False attribute if the vertex is
            vertices_types.append(2)
    elif builtins.current_renderer == "skia":
        vertex(x, y)


def bezier_vertex(x2, y2, x3, y3, x4, y4):
    """
    Specifies vertex coordinates for Bezier curves

    :param x2: x-coordinate of the first control point
    :type x2: float

    :param y2: y-coordinate of the first control point
    :type y2: float

    :param x3: x-coordinate of the second control point
    :type x3: float

    :param y3: y-coordinate of the second control point
    :type y3: float

    :param x4: x-coordinate of the anchor point
    :type x4: float

    :param y4: y-coordinate of the anchor point
    :type y4: float
    """
    global is_bezier
    is_bezier = True

    if p5.mode == "3D":
        return
    if builtins.current_renderer == "vispy":
        if is_contour:
            curr_contour_vertices.append((x2, y2, x3, y3, x4, y4))
            curr_contour_vertices_types.append(3)
        else:
            vertices.append((x2, y2, x3, y3, x4, y4))
            vertices_types.append(3)
    elif builtins.current_renderer == "skia":
        vert_data = [x2, y2, x3, y3, x4, y4, {"is_vert": False}]
        if is_contour:
            contour_vertices.append(vert_data)
        else:
            vertices.append(vert_data)


def quadratic_vertex(cx, cy, x3, y3):
    """
    Specifies vertex coordinates for quadratic Bezier curves

    :param cx: x-coordinate of the control point
    :type cx: float

    :param cy: y-coordinate of the control point
    :type cy: float

    :param x3: x-coordinate of the anchor point
    :type x3: float

    :param y3: y-coordinate of the anchor point
    :type y3: float

    """

    if p5.mode == "3D":
        return
    global is_quadratic
    is_quadratic = True

    if builtins.current_renderer == "vispy":
        if is_contour:
            curr_contour_vertices.append((cx, cy, x3, y3))
            curr_contour_vertices_types.append(4)
        else:
            vertices.append((cx, cy, x3, y3))
            vertices_types.append(3)
    elif builtins.current_renderer == "skia":
        vert_data = [cx, cy, x3, y3, {"is_vert": False}]
        if is_contour:
            contour_vertices.append(vert_data)
        else:
            vertices.append(vert_data)


def vertex(x, y, z=0):
    """
    All shapes are constructed by connecting a series of
    vertices. vertex() is used to specify the vertex
    coordinates for points, lines, triangles, quads,
    and polygons. It is used exclusively within the
    beginShape() and endShape() functions.

    :param x: x-coordinate of the vertex
    :type x: float

    :param y: y-coordinate of the vertex
    :type y: float

    :param z: z-coordinate of the vertex
    :type z: float
    """
    if p5.mode == "3D":
        return
    if builtins.current_renderer == "vispy":
        if is_contour:
            curr_contour_vertices.append((x, y, z))
            curr_contour_vertices_types.append(1)
        else:
            vertices.append((x, y, z))
            vertices_types.append(1)
    elif builtins.current_renderer == "skia":
        vert_data = [
            x,
            y,
            0,
            0,
            0,
            tuple(255 * c for c in p5.renderer.style.fill_color),
            tuple(255 * c for c in p5.renderer.style.stroke_color),
            {},
        ]
        vert_data[-1]["is_vert"] = True

        if is_contour:
            if len(contour_vertices) == 0:
                vert_data[-1]["move_to"] = True
            contour_vertices.append(vert_data)
        else:
            vertices.append(vert_data)


def begin_contour():
    """
    Use the beginContour() and endContour() functions
    to create negative shapes within shapes such as
    the center of the letter 'O'. beginContour() begins
    recording vertices for the shape and endContour() stops
    recording. The vertices that define a negative shape must
    "wind" in the opposite direction from the exterior shape.
    First draw vertices for the exterior clockwise order, then
    for internal shapes, draw vertices shape in counter-clockwise.

    """
    global is_contour, contour_vertices, contour_vertices_types, in_contour
    is_contour = True
    in_contour = True
    contour_vertices = []
    contour_vertices_types = []


def end_contour():
    """Ends the current contour.

    For more info, see :any:`begin_contour`.
    """
    global in_contour, curr_contour_vertices, curr_contour_vertices_types, is_first_contour
    in_contour = False
    # https://github.com/p5py/p5/pull/357#discussion_r935221732
    # is_contour = False
    if builtins.current_renderer == "vispy":
        # Close contour
        curr_contour_vertices.append(curr_contour_vertices[0])
        curr_contour_vertices_types.append(curr_contour_vertices_types[0])
        # Save contour
        contour_vertices.append(curr_contour_vertices)
        contour_vertices_types.append(curr_contour_vertices_types)
        curr_contour_vertices, curr_contour_vertices_types = [], []
    elif builtins.current_renderer == "skia":

        vert_data = copy.deepcopy(contour_vertices[0])
        vert_data[-1]["is_vert"] = contour_vertices[0][-1].get("is_vert", None)
        vert_data[-1]["move_to"] = False

        contour_vertices.append(vert_data)

        # Close the shape before starting the contour
        if is_first_contour:
            vertices.append(vertices[0])
            is_first_contour = False

        for vert in contour_vertices:
            vertices.append(vert)


def get_curve_vertices(verts):
    if len(verts) == 0:
        return []

    s = 1 - curves.curve_tightness_amount
    shape_vertices = [(verts[1][0], verts[1][1], 0.0)]
    steps = curves.curve_resolution

    for i in range(1, len(verts) - 2):
        v = verts[i]
        start = (
            shape_vertices[len(shape_vertices) - 1][0],
            shape_vertices[len(shape_vertices) - 1][1],
        )
        c1 = [
            v[0] + (s * verts[i + 1][0] - s * verts[i - 1][0]) / 6,
            v[1] + (s * verts[i + 1][1] - s * verts[i - 1][1]) / 6,
        ]
        c2 = [
            verts[i + 1][0] + (s * verts[i][0] - s * verts[i + 2][0]) / 6,
            verts[i + 1][1] + (s * verts[i][1] - s * verts[i + 2][1]) / 6,
        ]
        stop = [verts[i + 1][0], verts[i + 1][1]]

        for i in range(steps + 1):
            t = i / steps
            p = curves.bezier_point(start, c1, c2, stop, t)
            shape_vertices.append(Point(*p))

    return shape_vertices


def get_bezier_vertices(verts, vert_types):
    if len(verts) == 0:
        return []

    shape_vertices = []
    steps = curves.curve_resolution
    for i in range(len(verts)):
        if vert_types[i] == 1 or vert_types[i] == 2:
            shape_vertices.append((verts[i][0], verts[i][1], 0.0))
        else:
            start = (
                shape_vertices[len(shape_vertices) - 1][0],
                shape_vertices[len(shape_vertices) - 1][1],
            )
            c1 = [verts[i][0], verts[i][1]]
            c2 = [verts[i][2], verts[i][3]]
            stop = [verts[i][4], verts[i][5]]

            for i in range(steps + 1):
                t = i / steps
                p = curves.bezier_point(start, c1, c2, stop, t)
                shape_vertices.append(Point(*p))
    return shape_vertices


def get_quadratic_vertices(verts, vert_types):
    if len(verts) == 0:
        return []

    shape_vertices = []
    steps = curves.curve_resolution
    for i in range(len(verts)):
        if vert_types[i] == 1 or vert_types[i] == 2:
            shape_vertices.append((verts[i][0], verts[i][1], 0.0))
        else:
            start = (
                shape_vertices[len(shape_vertices) - 1][0],
                shape_vertices[len(shape_vertices) - 1][1],
            )
            control = [verts[i][0], verts[i][1]]
            stop = [verts[i][2], verts[i][3]]

            for i in range(steps + 1):
                t = i / steps
                p = curves.quadratic_point(start, control, stop, t)
                shape_vertices.append(Point(*p))

    return shape_vertices


def end_shape(mode=""):
    """
    The endShape() function is the companion to beginShape()
    and may only be called after beginShape(). When endshape()
    is called, all of image data defined since the previous call
    to beginShape() is rendered.

    :param mode: use CLOSE to close the shape
    :type mode: str

    """
    global is_bezier, is_curve, is_quadratic, is_contour, is_first_contour, in_contour
    if is_curve or is_bezier or is_quadratic:
        assert shape_kind == TESS, "Should not specify primitive type for a curve"
    assert not in_contour, "begin_contour called without calling end_contour"

    if len(vertices) == 0:
        return

    if (not p5.renderer.style.stroke_enabled) and (not p5.renderer.style.fill_enabled):
        return

    if builtins.current_renderer == "vispy":
        # if the shape is closed, the first element is also the last element
        if mode == "CLOSE":
            vertices.append(vertices[0])
            vertices_types.append(vertices_types[0])

        if is_curve:
            if len(vertices) > 3:
                p5.renderer.shape(
                    vertices=get_curve_vertices(vertices),
                    contours=[get_curve_vertices(c) for c in contour_vertices],
                    shape_type=TESS,
                )
        elif is_bezier:
            p5.renderer.shape(
                vertices=get_bezier_vertices(vertices, vertices_types),
                contours=[
                    get_bezier_vertices(contour_vertices[i], contour_vertices_types[i])
                    for i in range(len(contour_vertices))
                ],
                shape_type=TESS,
            )
        elif is_quadratic:
            p5.renderer.shape(
                vertices=get_quadratic_vertices(vertices, vertices_types),
                contours=[
                    get_quadratic_vertices(
                        contour_vertices[i], contour_vertices_types[i]
                    )
                    for i in range(len(contour_vertices))
                ],
                shape_type=TESS,
            )
        else:
            p5.renderer.shape(
                vertices=vertices, contours=contour_vertices, shape_type=shape_kind
            )

    elif builtins.current_renderer == "skia":
        close_shape = mode == "CLOSE"
        if close_shape and not is_contour:
            vertices.append(vertices[0])
        p5.renderer.end_shape(
            mode,
            vertices,
            is_curve,
            is_bezier,
            is_quadratic,
            is_contour,
            None if shape_kind == TESS else shape_kind,
        )
        if close_shape:
            vertices.pop()

    is_bezier = False
    is_curve = False
    is_quadratic = False
    is_contour = False
    is_first_contour = True



"""
This file hold utilities function for skia renderer
"""

from ...core import p5


def mode_adjust(a, b, c, d, mode):
    if mode == "CORNER":
        return {"x": a, "y": b, "w": c, "h": d}
    # TODO: Add documentation for CORNERS mode
    elif mode == "CORNERS":
        return {"x": a, "y": b, "w": c - a, "h": d - b}
    elif mode == "RADIUS":
        return {"x": a - c, "y": b - d, "w": c * 2, "h": d * 2}
    elif mode == "CENTER":
        return {"x": a - c * 0.5, "y": b - d * 0.5, "w": c, "h": d}
    else:
        raise ValueError(f"Unknown mode {mode}")


def should_draw():
    return p5.renderer.style.stroke_enabled or p5.renderer.style.fill_enabled



from . import p5
from ..sketch.util import ensure_p3d, scale_tuple
import numpy as np

__all__ = [
    "lights",
    "ambient_light",
    "directional_light",
    "point_light",
    "light_falloff",
    "light_specular",
]


def lights():
    """Sets the default ambient light, directional light, falloff, and specular values.

    The defaults are ambientLight(128, 128, 128), directionalLight(128, 128, 128, 0, 0, -1), and lightFalloff(1, 0, 0).
    """
    ambient_light(128, 128, 128)
    directional_light(128, 128, 128, 0, 0, -1)
    light_falloff(1, 0, 0)


def ambient_light(r, g, b):
    """Adds an ambient light.

    Ambient light comes from all directions towards all directions.
    Ambient lights are almost always used in combination with other types of lights.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float
    """
    ensure_p3d("ambient_light")
    p5.renderer.add_ambient_light(*scale_tuple((r, g, b)))


def directional_light(r, g, b, x, y, z):
    """Adds a directional light.
    Directional light comes from one direction: it is stronger when hitting a surface squarely,
    and weaker if it hits at a gentle angle.
    After hitting a surface, directional light scatters in all directions.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float

    :param x: x component of the direction vector
    :type x: float

    :param y: y component of the direction vector
    :type y: float

    :param z: z component of the direction vector
    :type z: float
    """
    ensure_p3d("directional_light")
    p5.renderer.add_directional_light(*scale_tuple((r, g, b)), x, y, z)


def point_light(r, g, b, x, y, z):
    """Adds a point light.
    Point light comes from one location and emits to all directions.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float

    :param x: x component of the location vector
    :type x: float

    :param y: y component of the location vector
    :type y: float

    :param z: z component of the location vector
    :type z: float
    """
    ensure_p3d("point_light")
    p5.renderer.add_point_light(*scale_tuple((r, g, b)), x, y, z)


def light_falloff(constant, linear, quadratic):
    """Sets the falloff rates for point lights. Affects only the elements which are created after it in the code.

    d = distance from light position to vertex position

    falloff = 1 / (constant + d * linear + (d*d) * quadratic)

    If the coefficient is 0, then that term is ignored.
    The P3D renderer defaults to (0, 0, 0), i.e. no falloff.

    :param constant: coefficient for the constant term
    :type constant: float

    :param linear: coefficient for the linear term
    :type linear: float

    :param quadratic: coefficient for the quadratic term
    :type quadratic: float
    """
    ensure_p3d("light_falloff")
    p5.renderer.curr_constant_falloff = constant
    p5.renderer.curr_linear_falloff = linear
    p5.renderer.curr_quadratic_falloff = quadratic


def light_specular(r, g, b):
    """Sets the specular color for lights. Only visible with :any:`p5.blinn_phong_material`. Is set to (0 ,0, 0) by default.
    Affects only the elements which are created after it in the code.

    Specular refers to light which bounces off a surface in a preferred direction
    (rather than bouncing in all directions like a diffuse light) and is used for
    creating highlights. The specular quality of a light interacts with the specular
    material qualities set through the `specular()` and `shininess()` functions.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float
    """
    ensure_p3d("light_specular")
    p5.renderer.light_specular = np.array(scale_tuple((r, g, b)))



from . import p5, fill
from ..sketch.util import ensure_p3d, scale_tuple
import numpy as np

__all__ = [
    "normal_material",
    "basic_material",
    "blinn_phong_material",
    "ambient",
    "emissive",
    "diffuse",
    "shininess",
    "specular",
]


class BasicMaterial:
    def __init__(self, color):
        self.color = color


class NormalMaterial:
    pass


class BlinnPhongMaterial:
    def __init__(self, ambient, diffuse, specular, shininess):
        self.ambient = ambient
        self.diffuse = diffuse
        self.specular = specular
        self.shininess = shininess


def normal_material():
    """The color is determined by the normal vector of a surface. Does not respond to lights.

    Useful for debugging.
    """
    ensure_p3d("normal_material")
    p5.renderer.style.material = NormalMaterial()


def basic_material(r, g, b):
    """The default material. Always displays a solid color.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float
    """
    ensure_p3d("basic_material")
    fill(r, g, b)
    p5.renderer.style.material = BasicMaterial(p5.renderer.style.fill_color)


def blinn_phong_material():
    """Material based on the Blinn-Phong reflection model. This is the most "realistic" material in p5py.

    Blinn-Phong shading can be decomposed into three parts:
    ambient, diffuse, and specular.

    The ambient component is essentially a constant term that is alway present.
    We calculate it by summing all the ambient lights in a scene and multiplying it
    with the normalized ambient coefficent set by ambient.

    The diffuse component takes the normal vector of a surface into account and
    varies how much light is reflected depending on the angle that the surface
    makes with the incoming light.

    The specular component not only accounts for the direction of the light
    (like the diffuse component) but also the direction of the viewer. If the
    viewer is not on the path of the reflected light, the specular component
    falls off quickly, producing the glossy reflections we see on some materials.

    The color shown on the user's screen is the sum of all three components.
    """
    ensure_p3d("blinn_phong_material")
    rend = p5.renderer.style
    p5.renderer.style.material = BlinnPhongMaterial(
        rend.ambient, rend.diffuse, rend.specular, rend.shininess
    )


def ambient(r, g, b):
    """Sets the ambient light color reflected by the next :any:`blinn_phong_material`.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float
    """
    ensure_p3d("ambient")
    p5.renderer.style.ambient = np.array(scale_tuple((r, g, b)), dtype=np.float32)


def emissive(r, g, b):
    """This function is the same as diffuse.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float
    """
    ensure_p3d("emissive")
    return diffuse(*scale_tuple((r, g, b)))


def diffuse(r, g, b):
    """Sets the diffuse light color reflected by the next :any:`blinn_phong_material`.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float
    """
    ensure_p3d("diffuse")
    p5.renderer.style.diffuse = np.array(scale_tuple((r, g, b)), dtype=np.float32)


def shininess(p):
    """Sets how glossy the next :any:`blinn_phong_material` is. This only affects the specular term.

    Should be used together with :any:`light_specular` and :any:`specular`.

    :param p: exponent of the cosine term in the `Blinn-Phong Reflection Model <https://en.wikipedia.org/wiki/Blinn%E2%80%93Phong_reflection_model>`_
    :type p: float
    """
    ensure_p3d("shininess")
    p5.renderer.style.shininess = p


def specular(r, g, b):
    """Sets the specular light color reflected by the next :any:`blinn_phong_material`.

    Should be used together with :any:`light_specular`.
    """
    ensure_p3d("specular")
    p5.renderer.style.specular = np.array(scale_tuple((r, g, b)), dtype=np.float32)


# TODO: Document default values for material functions in renderer3D



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import math
import builtins

from ..pmath import Point
from ..pmath import curves

from ..sketch.Vispy2DRenderer.shape import PShape
from .constants import ROUND, SQUARE, PROJECT

from . import p5

from ..sketch.Skia2DRenderer.util import should_draw, mode_adjust

__all__ = [
    "point",
    "line",
    "arc",
    "triangle",
    "quad",
    "rect",
    "square",
    "circle",
    "ellipse",
    "ellipse_mode",
    "rect_mode",
    "bezier",
    "curve",
    "create_shape",
]


def point(x, y, z=0):
    """Returns a point.

    :param x: x-coordinate of the shape.
    :type x: int or float

    :param y: y-coordinate of the shape.
    :type y: int or float

    :param z: z-coordinate of the shape (defaults to 0).
    :type z: int or float

    :returns: A point PShape.
    :rtype: PShape

    """
    if builtins.current_renderer == "vispy":
        if p5.renderer.style.stroke_cap == SQUARE:
            pass
        elif p5.renderer.style.stroke_cap == PROJECT:
            return square((x, y, z), p5.renderer.style.stroke_weight, mode="CENTER")
        elif p5.renderer.style.stroke_cap == ROUND:
            return circle((x, y, z), p5.renderer.style.stroke_weight / 2, mode="CENTER")
        raise ValueError("Unknown stroke_cap value")
    elif builtins.current_renderer == "skia":
        if p5.renderer.style.stroke_enabled:
            p5.renderer.point(x, y)


def line(*args):
    """Returns a line.

    :param x1: x-coordinate of the first point
    :type x1: float

    :param y1: y-coordinate of the first point
    :type y1: float

    :param z1: z-coordinate of the first point
    :type z1: float

    :param x2: x-coordinate of the first point
    :type x2: float

    :param y2: y-coordinate of the first point
    :type y2: float

    :param z2: z-coordinate of the first point
    :type z2: float

    :param p1: Coordinates of the starting point of the line.
    :type p1: tuple

    :param p2: Coordinates of the end point of the line.
    :type p2: tuple

    :returns: A line PShape.
    :rtype: PShape

    """
    if len(args) == 2:
        p1, p2 = args[0], args[1]
    elif len(args) == 4:
        p1, p2 = args[:2], args[2:]
    elif len(args) == 6:
        p1, p2 = args[:3], args[3:]
    else:
        raise ValueError("Unexpected number of arguments passed to line()")

    path = [Point(*p1), Point(*p2)]
    p5.renderer.line(path)


def bezier(*args):
    """Return a bezier path defined by two control points.

    :param x1: x-coordinate of the first anchor point
    :type x1: float

    :param y1: y-coordinate of the first anchor point
    :type y1: float

    :param z1: z-coordinate of the first anchor point
    :type z1: float

    :param x2: x-coordinate of the first control point
    :type x2: float

    :param y2: y-coordinate of the first control point
    :type y2: float

    :param z2: z-coordinate of the first control point
    :type z2: float

    :param x3: x-coordinate of the second control point
    :type x3: float

    :param y3: y-coordinate of the second control point
    :type y3: float

    :param z3: z-coordinate of the second control point
    :type z3: float

    :param x4: x-coordinate of the second anchor point
    :type x4: float

    :param y4: y-coordinate of the second anchor point
    :type y4: float

    :param z4: z-coordinate of the second anchor point
    :type z4: float

    :param start: The starting point of the bezier curve.
    :type start: tuple.

    :param control_point_1: The first control point of the bezier
        curve.
    :type control_point_1: tuple.

    :param control_point_2: The second control point of the bezier
        curve.
    :type control_point_2: tuple.

    :param stop: The end point of the bezier curve.
    :type stop: tuple.

    :returns: A bezier path.
    :rtype: PShape.

    """
    if len(args) == 4:
        start, control_point_1, control_point_2, stop = args
    elif len(args) == 8:
        start, control_point_1, control_point_2, stop = (
            args[:2],
            args[2:4],
            args[4:6],
            args[6:],
        )
    elif len(args) == 12:
        start, control_point_1, control_point_2, stop = (
            args[:3],
            args[3:6],
            args[6:9],
            args[9:],
        )
    else:
        raise ValueError("Unexpected number of arguments passed to bezier()")

    vertices = []
    steps = curves.bezier_resolution
    for i in range(steps + 1):
        t = i / steps
        p = curves.bezier_point(start, control_point_1, control_point_2, stop, t)
        vertices.append(p[:3])

    p5.renderer.bezier(vertices)


def curve(*args):
    """Return a Catmull-Rom curve defined by four points.

    :param x1: x-coordinate of the beginning control point
    :type x1: float

    :param y1: y-coordinate of the beginning control point
    :type y1: float

    :param z1: z-coordinate of the beginning control point
    :type z1: float

    :param x2: x-coordinate of the first point
    :type x2: float

    :param y2: y-coordinate of the first point
    :type y2: float

    :param z2: z-coordinate of the first point
    :type z2: float

    :param x3: x-coordinate of the second point
    :type x3: float

    :param y3: y-coordinate of the second point
    :type y3: float

    :param z3: z-coordinate of the second point
    :type z3: float

    :param x4: x-coordinate of the ending control point
    :type x4: float

    :param y4: y-coordinate of the ending control point
    :type y4: float

    :param z4: z-coordinate of the ending control point
    :type z4: float

    :param point_1: The first point of the curve.
    :type point_1: tuple

    :param point_2: The first point of the curve.
    :type point_2: tuple

    :param point_3: The first point of the curve.
    :type point_3: tuple

    :param point_4: The first point of the curve.
    :type point_4: tuple

    :returns: A curved path.
    :rtype: PShape

    """
    if len(args) == 4:
        point_1, point_2, point_3, point_4 = args
    elif len(args) == 8:
        point_1, point_2, point_3, point_4 = args[:2], args[2:4], args[4:6], args[6:]
    elif len(args) == 12:
        point_1, point_2, point_3, point_4 = args[:3], args[3:6], args[6:9], args[9:]
    else:
        raise ValueError("Unexpected number of arguments passed to curve()")

    vertices = []
    steps = curves.curve_resolution
    for i in range(steps + 1):
        t = i / steps
        p = curves.curve_point(point_1, point_2, point_3, point_4, t)
        vertices.append(p[:3])

    p5.renderer.curve(vertices)


def triangle(*args):
    """Return a triangle.

    :param x1: x-coordinate of the first point
    :type x1: float

    :param y1: y-coordinate of the first point
    :type y1: float

    :param x2: x-coordinate of the second point
    :type x2: float

    :param y2: y-coordinate of the second point
    :type y2: float

    :param x3: x-coordinate of the third point
    :type x3: float

    :param y3: y-coordinate of the third point
    :type y3: float

    :param p1: coordinates of the first point of the triangle
    :type p1: tuple | list | p5.Vector

    :param p2: coordinates of the second point of the triangle
    :type p2: tuple | list | p5.Vector

    :param p3: coordinates of the third point of the triangle
    :type p3: tuple | list | p5.Vector

    :returns: A triangle.
    :rtype: p5.PShape
    """
    if len(args) == 6:
        p1, p2, p3 = args[:2], args[2:4], args[4:]
    elif len(args) == 3:
        p1, p2, p3 = args
    else:
        raise ValueError("Unexpected number of arguments passed to triangle()")
    if builtins.current_renderer == "vispy":
        path = [Point(*p1), Point(*p2), Point(*p3)]
        p5.renderer.triangle(path)
    elif builtins.current_renderer == "skia":
        if should_draw():
            p5.renderer.triangle(*p1, *p2, *p3)


def quad(*args):
    """Return a quad.

    :param x1: x-coordinate of the first point
    :type x1: float

    :param y1: y-coordinate of the first point
    :type y1: float

    :param x2: x-coordinate of the second point
    :type x2: float

    :param y2: y-coordinate of the second point
    :type y2: float

    :param x3: x-coordinate of the third point
    :type x3: float

    :param y3: y-coordinate of the third point
    :type y3: float

    :param x4: x-coordinate of the forth point
    :type x4: float

    :param y4: y-coordinate of the forth point
    :type y4: float

    :param p1: coordinates of the first point of the quad
    :type p1: tuple | list | p5.Vector

    :param p2: coordinates of the second point of the quad
    :type p2: tuple | list | p5.Vector

    :param p3: coordinates of the third point of the quad
    :type p3: tuple | list | p5.Vector

    :param p4: coordinates of the fourth point of the quad
    :type p4: tuple | list | p5.Vector

    :returns: A quad.
    :rtype: PShape
    """
    if len(args) == 8:
        p1, p2, p3, p4 = args[:2], args[2:4], args[4:6], args[6:]
    elif len(args) == 4:
        p1, p2, p3, p4 = args
    else:
        raise ValueError("Unexpected number of arguments passed to quad()")

    if builtins.current_renderer == "vispy":
        path = [Point(*p1), Point(*p2), Point(*p3), Point(*p4)]
        p5.renderer.quad(path)
    elif builtins.current_renderer == "skia":
        x1, y1 = p1
        x2, y2 = p2
        x3, y3 = p3
        x4, y4 = p4
        p5.renderer.quad(x1, y1, x2, y2, x3, y3, x4, y4)


def rect(*args, mode=None):
    """Return a rectangle.

    :param x: x-coordinate of the rectangle by default
    :type float:

    :param y: y-coordinate of the rectangle by default
    :type float:

    :param w: width of the rectangle by default
    :type float:

    :param h: height of the rectangle by default
    :type float:

    :param coordinate: Represents the lower left corner of then
        rectangle when mode is 'CORNER', the center of the rectangle
        when mode is 'CENTER' or 'RADIUS', and an arbitrary corner
        when mode is 'CORNERS'

    :type coordinate: tuple | list | p5.Vector

    :param args: For modes'CORNER' or 'CENTER' this has the form
        (width, height); for the 'RADIUS' this has the form
        (half_width, half_height); and for the 'CORNERS' mode, args
        should be the corner opposite to `coordinate`.

    :type: tuple

    :param mode: The drawing mode for the rectangle. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'} (defaults to the
        mode being used by the p5.renderer.)

    :type mode: str

    :returns: A rectangle.
    :rtype: p5.PShape

    TODO: Update docs for rect, we support border radius as well
    """
    if builtins.current_renderer == "vispy":
        if len(args) == 4:
            coordinate, args = args[:2], args[2:]
        elif len(args) == 3:
            coordinate, args = args[0], args[1:]
        else:
            raise ValueError("Unexpected number of arguments passed to rect()")

        if mode is None:
            mode = p5.renderer.style.rect_mode

        if mode == "CORNER":
            corner = coordinate
            width, height = args
        elif mode == "CENTER":
            center = Point(*coordinate)
            width, height = args
            corner = Point(center.x - width / 2, center.y - height / 2, center.z)
        elif mode == "RADIUS":
            center = Point(*coordinate)
            half_width, half_height = args
            corner = Point(center.x - half_width, center.y - half_height, center.z)
            width = 2 * half_width
            height = 2 * half_height
        elif mode == "CORNERS":
            corner = Point(*coordinate)
            (corner_2,) = args
            corner_2 = Point(*corner_2)
            width = corner_2.x - corner.x
            height = corner_2.y - corner.y
        else:
            raise ValueError("Unknown rect mode {}".format(mode))

        p1 = Point(*corner)
        p2 = Point(p1.x + width, p1.y, p1.z)
        p3 = Point(p2.x, p2.y + height, p2.z)
        p4 = Point(p1.x, p3.y, p3.z)
        return quad(p1, p2, p3, p4)

    elif builtins.current_renderer == "skia":
        if len(args) == 4:
            x, y, w, h = args
        elif len(args) == 3:
            x, y = args[0]
            w, h = args[1:]
        vals = mode_adjust(
            x, y, w, h, mode if mode is not None else p5.renderer.style.rect_mode
        )
        p5.renderer.rect(*(vals["x"], vals["y"], vals["w"], vals["h"]) + args[4:])


def square(*args, mode=None):
    """Return a square.

    :param x: x-coordinate of the square by default
    :type float:

    :param y: y-coordinate of the square by default
    :type float:

    :param coordinate: When mode is set to 'CORNER', the coordinate
        represents the lower-left corner of the square. For modes
        'CENTER' and 'RADIUS' the coordinate represents the center of
        the square.

    :type coordinate: tuple | list | p5.Vector

    :param side_length: The side_length of the square (for modes
        'CORNER' and 'CENTER') or hald of the side length (for the
        'RADIUS' mode)

    :type side_length: int or float

    :param mode: The drawing mode for the square. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'} (defaults to the
        mode being used by the p5.renderer.)

    :type mode: str

    :returns: A rectangle.
    :rtype: p5.PShape

    :raises ValueError: When the mode is set to 'CORNERS'

    """
    if builtins.current_renderer == "vispy":
        if len(args) == 2:
            coordinate, side_length = args
        elif len(args) == 3:
            coordinate, side_length = args[:2], args[2]
        else:
            raise ValueError("Unexpected number of arguments passed to square()")

        if mode is None:
            mode = p5.renderer.style.rect_mode

        if mode == "CORNERS":
            raise ValueError("Cannot draw square with {} mode".format(mode))
        return rect(coordinate, side_length, side_length, mode=mode)
    elif builtins.current_renderer == "skia":
        if should_draw():
            if len(args) == 2:
                x, y = args[0]
                side = args[1]
            elif len(args) == 3:
                x, y, side = args
            rect(x, y, side, side)


def rect_mode(mode="CORNER"):
    """Change the rect and square drawing mode for the p5.renderer.

    :param mode: The new mode for drawing rects. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'}. This defaults to
        'CORNER' so calling rect_mode without parameters will reset
        the sketch's rect mode.
    :type mode: str

    """
    p5.renderer.style.rect_mode = mode


def arc(*args, mode=None, ellipse_mode=None):
    """Return a ellipse.

    :param x: x-coordinate of the arc's ellipse.
    :type x: float

    :param y: y-coordinate of the arc's ellipse.
    :type y: float

    :param coordinate: Represents the center of the arc when mode
        is 'CENTER' (the default) or 'RADIUS', the lower-left corner
        of the ellipse when mode is 'CORNER'.

    :type coordinate: 3-tuple

    :param width: For ellipse modes 'CORNER' or 'CENTER' this
        represents the width of the the ellipse of which the arc is a
        part. Represents the x-radius of the parent ellipse when
        ellipse mode is 'RADIUS

    :type width: float

    :param height: For ellipse modes 'CORNER' or 'CENTER' this
        represents the height of the the ellipse of which the arc is a
        part. Represents the y-radius of the parent ellipse when
        ellipse mode is 'RADIUS

    :type height: float

    :param mode: The mode used to draw an arc can be any of {None, 'OPEN', 'CHORD', 'PIE'}.

    :type mode: str

    :param ellipse_mode: The drawing mode used for the ellipse. Should be one of
        {'CORNER', 'CENTER', 'RADIUS'} (defaults to the
        mode being used by the p5.renderer.)

    :type mode: str

    :returns: An arc.
    :rtype: Arc

    """
    if len(args) == 5:
        coordinate, width, height, start_angle, stop_angle = args
    elif len(args) == 6:
        coordinate = args[:2]
        width, height, start_angle, stop_angle = args[2:]
    else:
        raise ValueError("Unexpected number of arguments passed to arc()")

    if ellipse_mode is None:
        emode = p5.renderer.style.ellipse_mode
    else:
        emode = ellipse_mode

    if builtins.current_renderer == "vispy":
        if emode == "CORNER":
            corner = Point(*coordinate)
            dim = Point(width / 2, height / 2)
            center = (corner.x + dim.x, corner.y + dim.y, corner.z)
        elif emode == "CENTER":
            center = Point(*coordinate)
            dim = Point(width / 2, height / 2)
        elif emode == "RADIUS":
            center = Point(*coordinate)
            dim = Point(width, height)
        else:
            raise ValueError("Unknown arc mode {}".format(emode))
        p5.renderer.arc(center, dim, start_angle, stop_angle, mode)

    elif builtins.current_renderer == "skia":
        if not should_draw() or start_angle == stop_angle:
            return
        x, y = coordinate[0], coordinate[1]
        vals = mode_adjust(x, y, width, height, emode)
        # TODO: normalize_arc_angles
        p5.renderer.arc(
            vals["x"], vals["y"], vals["w"], vals["h"], start_angle, stop_angle, mode
        )


def ellipse(*args, mode=None):
    """Return a ellipse.

    :param a: x-coordinate of the ellipse
    :type a: float

    :param b: y-coordinate of the ellipse
    :type b: float

    :param c: width of the ellipse
    :type c: float

    :param d: height of the ellipse
    :type d: float

    :param coordinate: Represents the center of the ellipse when mode
        is 'CENTER' (the default) or 'RADIUS', the lower-left corner
        of the ellipse when mode is 'CORNER' or, and an arbitrary
        corner when mode is 'CORNERS'.

    :type coordinate: 3-tuple

    :param args: For modes'CORNER' or 'CENTER' this has the form
        (width, height); for the 'RADIUS' this has the form
        (x_radius, y_radius); and for the 'CORNERS' mode, args
        should be the corner opposite to `coordinate`.

    :type: tuple

    :param mode: The drawing mode for the ellipse. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'} (defaults to the
        mode being used by the p5.renderer.)

    :type mode: str

    :returns: An ellipse
    :rtype: Arc

    """
    if len(args) == 3:
        coordinate, args = args[0], args[1:]
    elif len(args) == 4:
        coordinate, args = args[:2], args[2:]
    else:
        raise ValueError("Unexpected number of arguments passed to ellipse()")

    if mode is None:
        mode = p5.renderer.style.ellipse_mode

    if builtins.current_renderer == "vispy":
        if mode == "CORNERS":
            corner = Point(*coordinate)
            (corner_2,) = args
            corner_2 = Point(*corner_2)
            width = corner_2.x - corner.x
            height = corner_2.y - corner.y
            mode = "CORNER"
        else:
            width, height = args
        return arc(
            coordinate, width, height, 0, math.pi * 2, mode="CHORD", ellipse_mode=mode
        )

    elif builtins.current_renderer == "skia":
        if not should_draw():
            return
        x, y = coordinate
        w, h = args
        vals = mode_adjust(x, y, w, h, mode)
        p5.renderer.ellipse(vals["x"], vals["y"], vals["w"], vals["h"])


def circle(*args, mode=None):
    """Return a circle.

    :param x: x-coordinate of the centre of the circle.
    :type x: float

    :param y: y-coordinate of the centre of the circle.
    :type y: float

    :param coordinate: Represents the center of the ellipse when mode
        is 'CENTER' (the default) or 'RADIUS', the lower-left corner
        of the ellipse when mode is 'CORNER' or, and an arbitrary
        corner when mode is 'CORNERS'.

    :type coordinate: 3-tuple

    :param diameter: For modes'CORNER' or 'CENTER' this actually
        represents the diameter; for the 'RADIUS' this represents the
        radius.

    :type diameter: float

    :param mode: The drawing mode for the ellipse. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'} (defaults to the
        mode being used by the p5.renderer.)

    :type mode: str

    :returns: A circle.
    :rtype: Ellipse

    :raises ValueError: When mode is set to 'CORNERS'

    """
    if len(args) == 2:
        coordinate, diameter = args[0], args[1]
    elif len(args) == 3:
        coordinate, diameter = args[:2], args[2]
    else:
        raise ValueError("Unexpected number of arguments passed to circle()")

    if mode is None:
        mode = p5.renderer.style.ellipse_mode
    if builtins.current_renderer == "vispy":
        if mode == "CORNERS":
            raise ValueError("Cannot create circle in CORNERS mode")
        return ellipse(coordinate, diameter, diameter, mode=mode)
    elif builtins.current_renderer == "skia":
        x, y = coordinate
        p5.renderer.circle(x, y, diameter)


def ellipse_mode(mode="CENTER"):
    """Change the ellipse and circle drawing mode for the p5.renderer.

    :param mode: The new mode for drawing ellipses. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'}. This defaults to
        'CENTER' so calling ellipse_mode without parameters will reset
        the sketch's ellipse mode.
    :type mode: str

    """
    p5.renderer.style.ellipse_mode = mode


def create_shape(kind=None, *args, **kwargs):
    """Create a new PShape

    Note :: A shape created using this function is *not* visible by
        default. Please make the shape visible by setting the shapes's
        `visible` attribute to true.

    :param kind: Type of shape. When left unspecified a generic PShape
        is returned (which can be edited later). Valid values for
        `kind` are: { 'point', 'line', 'triangle', 'quad', 'rect',
        'ellipse', 'arc', }

    :type kind: None | str

    :param args: Initial arguments to be passed to the shape creation
        function (only applied when `kind` is *not* None)

    :param kwargs: Initial keyword arguments to be passed to the shape
        creation function (only applied when `kind` is *not* None)

    :returns: The requested shape
    :rtype: p5.PShape

    """

    # TODO: add 'box', 'sphere' support
    valid_values = {
        None,
        "point",
        "line",
        "triangle",
        "quad",
        "rect",
        "square",
        "ellipse",
        "circle",
        "arc",
    }

    def empty_shape(*args, **kwargs):
        return PShape()

    shape_map = {
        "arc": arc,
        "circle": circle,
        "ellipse": ellipse,
        "line": line,
        "point": point,
        "quad": quad,
        "rect": rect,
        "square": square,
        "triangle": triangle,
        None: empty_shape,
    }

    # kwargs['visible'] = False
    return shape_map[kind](*args, **kwargs)



import builtins
from p5.core import p5

import contextlib, glfw, skia
from OpenGL import GL
from time import time

from ..events import handler_names
from .handlers import *
from .util import *


def _dummy(*args, **kwargs):
    """Eat all arguments, do nothing."""
    pass


class SkiaSketch:
    def __init__(self, setup_method, draw_method, handlers=dict(), frame_rate=60):
        self._size = (600, 400)
        self.setup_method = setup_method
        self.draw_method = draw_method

        self.surface = None
        self.context = None
        self.window = None
        self.canvas = None

        self.main_loop_state = True
        self.looping = True
        self.redraw = False

        self.paint = skia.Paint()
        self.paint.setAntiAlias(True)
        self.path = skia.Path()

        self.frame_rate = frame_rate

        """
        resized : (boolean) 
        When we make a call to resize the window, it does not happen instantly
        but rather glfw places the resize call on a different thread and later resizes the window 
        and its framebuffer triggering their callbacks.
        This boolean variable is used to keep track whether a resize call was made and the 
        window is yet to be resized. If so, halt the rendering process and wait for the window to
        be resized.
        This is necessary to ensure correct no_loop() behaviour.
        """
        self.resized = True

        # we can discard using builtins, by using p5.variableName
        builtins.frame_count = 0

        self.handlers = {}

        for handler_name in handler_names:
            self.handlers[handler_name] = handlers.get(handler_name, _dummy)

        self.handler_queue = []

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, val):
        self._size = val
        self.resize()

    def clean_up(self):
        glfw.terminate()
        self.context.abandonContext()

    def glfw_window(self):
        if not glfw.init():
            raise RuntimeError("glfw.init() failed")
        glfw.window_hint(glfw.STENCIL_BITS, 8)
        glfw.window_hint(glfw.SAMPLES, 4)
        GL.glEnable(GL.GL_MULTISAMPLE)
        window = glfw.create_window(*self._size, "p5py", None, None)
        glfw.make_context_current(window)
        return window

    def skia_surface(self, window, size):
        self.context = skia.GrDirectContext.MakeGL()
        backend_render_target = skia.GrBackendRenderTarget(
            *size,
            0,  # sampleCnt
            0,  # stencilBits
            skia.GrGLFramebufferInfo(0, GL.GL_RGBA8)
        )
        surface = skia.Surface.MakeFromBackendRenderTarget(
            self.context,
            backend_render_target,
            skia.kBottomLeft_GrSurfaceOrigin,
            skia.kRGBA_8888_ColorType,
            skia.ColorSpace.MakeSRGB(),
        )
        assert surface is not None
        return surface

    # create a new surface everytime
    def create_surface(self, size=None):
        if not size:
            size = self._size
        self._size = size
        builtins.width, builtins.height = size
        self.surface = self.skia_surface(self.window, size)
        self.canvas = self.surface.getCanvas()
        p5.renderer.initialize_renderer(self.canvas, self.paint, self.path)

    def main_loop(self):
        last_render_call_time = 0
        # Before starting the main while loop, check whether no_loop is called
        # If called we have to render a frame once
        # Set redraw and looping to False
        # This is done to ensure draw() is called atleast once

        while self.main_loop_state:
            if (
                self.resized
                and (self.looping or self.redraw)
                and (time() - last_render_call_time) > 1 / self.frame_rate
            ):
                builtins.frame_count += 1
                with self.surface as self.canvas:
                    self.draw_method()
                self.surface.flushAndSubmit()
                glfw.swap_buffers(self.window)
                last_render_call_time = time()

                # If redraw == True, we have rendered the frame once
                # Now don't render the next one
                if self.redraw:
                    self.redraw = False

                # Reset every style values back to default
                # TODO: Find a way to reset values of Graphic objects as well,
                # TODO: we can probably emit event after each loop to notify all graphics object
                p5.renderer.reset()

            self.poll_events()
            while len(self.handler_queue) != 0:
                function, event = self.handler_queue.pop(0)
                event._update_builtins()
                function(event)

    def start(self):
        self.window = self.glfw_window()
        self.create_surface()
        self.assign_callbacks()

        p5.renderer.initialize_renderer(self.canvas, self.paint, self.path)
        self.setup_method()
        self.poll_events()
        p5.renderer.render()
        self.surface.flushAndSubmit()
        glfw.swap_buffers(self.window)

        self.main_loop()
        self.clean_up()

    def resize(self):
        # call change the window size(), this will not be done instantly
        # but after some time and a frame_buffer_changed callback will occur on
        # on a different thread
        glfw.set_window_size(self.window, *self.size)

        # when glfw changes the framebuffer size, we will be resized completely
        # until then hold the rendering calls
        self.resized = False

    def poll_events(self):
        glfw.poll_events()
        if glfw.get_key(
            self.window, glfw.KEY_ESCAPE
        ) == glfw.PRESS or glfw.window_should_close(self.window):
            glfw.set_window_should_close(self.window, 1)
            self.main_loop_state = False

    # Callbacks and handlers
    def assign_callbacks(self):
        glfw.set_framebuffer_size_callback(
            self.window, self.frame_buffer_resize_callback_handler
        )

        glfw.set_key_callback(self.window, on_key_press)
        glfw.set_char_callback(self.window, on_key_char)
        glfw.set_mouse_button_callback(self.window, on_mouse_button)
        glfw.set_scroll_callback(self.window, on_mouse_scroll)
        glfw.set_cursor_pos_callback(self.window, on_mouse_motion)
        glfw.set_window_close_callback(self.window, on_close)
        glfw.set_window_focus_callback(self.window, on_window_focus)

    def frame_buffer_resize_callback_handler(self, window, width, height):
        """
        Gets called whenever frame buffer resizes
        Values of width and height may not be equal to the actual window's width and height
        in Retina Display
        """

        # Callback handler for frame buffer resize events
        GL.glViewport(0, 0, width, height)
        self.create_surface(size=(width, height))
        # with self.surface as self.canvas:
        #     # redraw on the canvas/ ( new frame buffer ) after resizing
        #     # and do not rewind/clear the path

        # Tell the program, we have resized the frame buffer
        # Restart the rendering again
        self.resized = True

    def _enqueue_event(self, handler_name, event):
        self.handler_queue.append((self.handlers[handler_name], event))



from abc import ABC
import numpy as np

from p5.core import p5
from p5.core.constants import SType, ROUND, MITER
from p5.pmath import matrix
from .shape import Arc, PShape

from dataclasses import dataclass
from vispy.gloo import Program, VertexBuffer, FrameBuffer, IndexBuffer
from OpenGL.GLU import (
    gluTessBeginPolygon,
    gluTessBeginContour,
    gluTessEndPolygon,
    gluTessEndContour,
    gluTessVertex,
)

# Useful constants
COLOR_WHITE = (1, 1, 1, 1)
COLOR_BLACK = (0, 0, 0, 1)


def to_3x3(mat):
    """Returns the upper left 3x3 corner of an np.array"""
    return mat[:3, :3]


def _tess_new_contour(vertices):
    """Given a list of vertices, evoke gluTess to create a contour"""
    gluTessBeginContour(p5.tess.tess)
    for v in vertices:
        gluTessVertex(p5.tess.tess, v, v)
    gluTessEndContour(p5.tess.tess)


def _vertices_to_render_primitive(gl_name, vertices):
    """Returns a render primitive of gl_type with vertices in sequential order"""
    return [gl_name, np.asarray(vertices), np.arange(len(vertices), dtype=np.uint32)]


def _get_line_from_verts(vertices):
    """Given a list of vertices, chain them sequentially in a line rendering primitive"""
    n_vert = len(vertices)
    return _get_line_from_indices(
        vertices,
        np.arange(n_vert - 1, dtype=np.uint32),
        np.arange(1, n_vert, dtype=np.uint32),
    )


def _get_line_from_indices(vertices, start, end):
    """Given two columns of indices that represent edges, return a line rendering primitive

    :param vertices: List of vertices
    :type vertices: list

    :param start: Array of start positions of edges in vertex indices
    :type start: np.ndarray

    :param end: Array of end positions fo edges in vertex indices
    :type end: np.ndarray
    """
    start = np.asarray(start, dtype=np.uint32)
    end = np.asarray(end, dtype=np.uint32)
    return [
        "lines",
        np.asarray(vertices),
        np.hstack((np.vstack(start), np.vstack(end))),
    ]


def _add_edges_to_primitive_list(primitive_list, vertices, start, end):
    """Adds edges to a list of render primitives, given their start and end positions (in vertex indices)

    :param start: Array of start positions of edges in vertex indices
    :type start: np.ndarray

    :param end: Array of end positions fo edges in vertex indices
    :type end: np.ndarray
    """
    primitive_list.append(_get_line_from_indices(vertices, start, end))


def _not_enough_vertices(shape, n):
    """Returns an error string that describes how many vertices are needed"""
    return "Need at least {} vertices for {}".format(n, shape)


def _wrong_multiple(shape, n):
    """Returns an error string that describes the # of vertices is not a multiple of n"""
    return "{} requires the number of vertices to be a multiple of {}".format(
        shape.shape_type, n
    )


def _check_shape(shape):
    """Checks if the shape is valid using assertions"""
    n_vert = len(shape.vertices)
    if shape.shape_type in [SType.TRIANGLES, SType.TRIANGLE_FAN, SType.TRIANGLE_STRIP]:
        assert n_vert >= 3, _not_enough_vertices(shape, 3)
    elif shape.shape_type in [SType.LINES, SType.LINE_STRIP]:
        assert n_vert >= 2, _not_enough_vertices(shape, 2)
    elif shape.shape_type in [SType.QUADS, SType.QUAD_STRIP]:
        assert n_vert >= 4, _not_enough_vertices(shape, 4)

    if shape.shape_type == SType.TRIANGLES:
        assert n_vert % 3 == 0, _wrong_multiple(shape, 3)
    if shape.shape_type == SType.QUADS:
        assert n_vert % 4 == 0, _wrong_multiple(shape, 4)


def _get_borders(shape):
    """Generates the render primitives for the borders of a given shape

    :returns: ['lines', vertices, idx]
    """
    render_primitives = []
    n_vert = len(shape.vertices)
    if shape.shape_type == SType.TRIANGLES:
        start = np.arange(n_vert)
        end = np.arange(n_vert) + np.tile([1, 1, -2], n_vert // 3)
        _add_edges_to_primitive_list(render_primitives, shape.vertices, start, end)
    elif shape.shape_type == SType.TRIANGLE_STRIP:
        start = np.concatenate((np.arange(n_vert - 1), np.arange(n_vert - 2)))
        end = np.concatenate((np.arange(1, n_vert), np.arange(2, n_vert)))
        _add_edges_to_primitive_list(render_primitives, shape.vertices, start, end)
    elif shape.shape_type == SType.TRIANGLE_FAN:
        start = np.concatenate((np.repeat([0], n_vert - 1), np.arange(1, n_vert - 1)))
        end = np.concatenate((np.arange(1, n_vert), np.arange(2, n_vert)))
        _add_edges_to_primitive_list(render_primitives, shape.vertices, start, end)
    elif shape.shape_type == SType.QUADS:
        start = np.arange(n_vert)
        end = np.arange(n_vert) + np.tile([1, 1, 1, -3], n_vert // 4)
        _add_edges_to_primitive_list(render_primitives, shape.vertices, start, end)
    elif shape.shape_type == SType.QUAD_STRIP:
        start = np.concatenate((np.arange(0, n_vert, 2), np.arange(n_vert - 2)))
        end = np.concatenate((np.arange(1, n_vert, 2), np.arange(2, n_vert)))
        _add_edges_to_primitive_list(render_primitives, shape.vertices, start, end)
    elif shape.shape_type == SType.LINES:
        start = np.arange(0, n_vert, 2)
        end = np.arange(1, n_vert, 2)
        _add_edges_to_primitive_list(render_primitives, shape.vertices, start, end)
    elif shape.shape_type == SType.LINE_STRIP:
        render_primitives.append(_get_line_from_verts(shape.vertices))
    elif shape.shape_type == SType.TESS:
        render_primitives.append(_get_line_from_verts(shape.vertices))
        for contour in shape.contours:
            render_primitives.append(_get_line_from_verts(contour))
    return render_primitives


def _get_meshes(shape):
    """Generates the rendering primitives for the meshes of a given shape

    :returns: [shape_type, vertices, idx]
    """
    render_primitives = []
    n_vert = len(shape.vertices)
    if shape.shape_type in [
        SType.TRIANGLES,
        SType.TRIANGLE_STRIP,
        SType.TRIANGLE_FAN,
        SType.QUAD_STRIP,
    ]:
        gl_name = shape.shape_type.name.lower()
        if gl_name == "quad_strip":  # vispy does not support quad_strip
            gl_name = "triangle_strip"  # but it can be drawn using triangle_strip
        render_primitives.append(_vertices_to_render_primitive(gl_name, shape.vertices))
    elif shape.shape_type == SType.QUADS:
        n_quad = len(shape.vertices) // 4
        render_primitives.append(
            [
                "triangles",
                np.asarray(shape.vertices),
                np.repeat(np.arange(0, n_vert, 4, dtype=np.uint32), 6)
                + np.tile(np.array([0, 1, 2, 0, 2, 3], dtype=np.uint32), n_quad),
            ]
        )
    elif shape.shape_type == SType.TESS:
        gluTessBeginPolygon(p5.tess.tess, None)
        _tess_new_contour(shape.vertices)
        if len(shape.contours) > 0:
            for contour in shape.contours:
                _tess_new_contour(contour)
        gluTessEndPolygon(p5.tess.tess)
        render_primitives += p5.tess.get_result()
    return render_primitives


def get_render_primitives(shape):
    """Given a shape, return a list of render primitives in the form of [type, vertices, indices]"""
    _check_shape(shape)
    render_primitives = []
    if isinstance(shape, Arc):
        # Render meshes
        if p5.renderer.style.fill_enabled:
            render_primitives.extend(_get_meshes(shape))
        # Render borders
        if p5.renderer.style.stroke_enabled:
            if shape.arc_mode in ["CHORD", "OPEN"]:  # Implies shape.shape_type == TESS
                render_primitives.extend(_get_borders(shape))
            elif shape.arc_mode is None:  # Implies shape.shape_type == TRIANGLE_FAN
                render_primitives.append(_get_line_from_verts(shape.vertices[1:]))
            elif shape.arc_mode == "PIE":  # Implies shape.shape_type == TRIANGLE_FAN
                render_primitives.append(_get_line_from_verts(shape.vertices))
    else:
        # Render points
        if shape.shape_type == SType.POINTS:
            render_primitives.append(
                _vertices_to_render_primitive(render_primitives, "points")
            )
        # Render meshes
        if p5.renderer.style.fill_enabled:
            render_primitives.extend(_get_meshes(shape))
        # Render borders
        if p5.renderer.style.stroke_enabled:
            render_primitives.extend(_get_borders(shape))
    return render_primitives


@dataclass
class Style2D:
    background_color = (0.8, 0.8, 0.8, 1.0)
    fill_color = COLOR_WHITE
    fill_enabled = True
    stroke_color = COLOR_BLACK
    stroke_enabled = True
    stroke_weight = 1
    tint_color = COLOR_BLACK
    tint_enabled = False
    ellipse_mode = "CENTER"
    rect_mode = "CORNER"
    color_parse_mode = "RGB"
    color_range = (255, 255, 255, 255)
    stroke_cap = ROUND
    stroke_join = MITER

    def set_stroke_cap(self, c):
        self.stroke_cap = c

    def set_stroke_join(self, j):
        self.stroke_join = j


# Abstract class that contains common code for OpenGL renderers
class OpenGLRenderer(ABC):
    def __init__(self, src_fbuffer, src_default):
        self.fbuffer_prog = Program(src_fbuffer.vert, src_fbuffer.frag)
        self.default_prog = Program(src_default.vert, src_default.frag)

        self.fbuffer = FrameBuffer()
        self.fbuffer_tex_front = None
        self.fbuffer_tex_back = None

        vertices = np.array(
            [[-1.0, -1.0], [+1.0, -1.0], [-1.0, +1.0], [+1.0, +1.0]], np.float32
        )
        texcoords = np.array(
            [[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [1.0, 1.0]], dtype=np.float32
        )

        self.fbuf_vertices = VertexBuffer(data=vertices)
        self.fbuf_texcoords = VertexBuffer(data=texcoords)

        self.fbuffer_prog["texcoord"] = self.fbuf_texcoords
        self.fbuffer_prog["position"] = self.fbuf_vertices

        self.vertex_buffer = VertexBuffer()
        self.index_buffer = IndexBuffer()

        # Renderer Globals
        # VIEW MATRICES, ETC
        #
        self.viewport = None
        self.texture_viewport = None
        self.transform_matrix = np.identity(4)
        self.projection_matrix = np.identity(4)

        # Renderer Globals: RENDERING
        self.draw_queue = []

        self.style = Style2D()
        self.style_stack = []
        self.matrix_stack = []

    def render_default(self, draw_type, draw_queue):
        # 1. Get the maximum number of vertices persent in the shapes
        # in the draw queue.
        #
        if len(draw_queue) == 0:
            return

        num_vertices = 0
        for vertices, _, _ in draw_queue:
            num_vertices = num_vertices + len(vertices)

        # 2. Create empty buffers based on the number of vertices.
        #
        data = np.zeros(
            num_vertices, dtype=[("position", np.float32, 3), ("color", np.float32, 4)]
        )

        # 3. Loop through all the shapes in the geometry queue adding
        # it's information to the buffer.
        #
        sidx = 0
        draw_indices = []
        for vertices, idx, color in draw_queue:
            num_shape_verts = len(vertices)

            data["position"][
                sidx : (sidx + num_shape_verts),
            ] = np.array(vertices)

            color_array = np.array([color] * num_shape_verts)
            data["color"][sidx : sidx + num_shape_verts, :] = color_array

            draw_indices.append(sidx + idx)

            sidx += num_shape_verts

        self.vertex_buffer.set_data(data)
        self.index_buffer.set_data(np.hstack(draw_indices))

        # 4. Bind the buffer to the shader.
        #
        self.default_prog.bind(self.vertex_buffer)

        # 5. Draw the shape using the proper shape type and get rid of
        # the buffers.
        #
        self.default_prog.draw(draw_type, indices=self.index_buffer)

    def cleanup(self):
        """Run the clean-up routine for the renderer.

        This method is called when all drawing has been completed and the
        program is about to exit.

        """
        self.default_prog.delete()
        self.fbuffer_prog.delete()
        self.fbuffer.delete()

    def _transform_vertices(self, vertices, local_matrix, global_matrix):
        """Applies `local_matrix` then `global_matrix` to `vertices`"""
        product = np.dot(np.dot(vertices, local_matrix.T), global_matrix.T)
        # dehomogenize coordinates
        # need np.newaxis to broadcast the vector because each row represents a vertex
        dehomogenized = product / product[:, 3][:, np.newaxis]
        return dehomogenized[:, :3]  # Return the first three rows of the result

    def push_matrix(self):
        """Pushes the current transformation matrix onto the matrix stack."""
        self.matrix_stack.append(self.transform_matrix.copy())

    def pop_matrix(self):
        """Pops the current transformation matrix off the matrix stack."""
        assert len(self.matrix_stack) > 0, "No matrix to pop"
        self.transform_matrix = self.matrix_stack.pop()

    def scale(self, sx, sy=None, sz=None):
        if sy is None and sz is None:
            sy = sx
            sz = sx
        elif sz is None:
            sz = 1
        tmat = matrix.scale_transform(sx, sy, sz)
        self.transform_matrix = self.transform_matrix.dot(tmat)
        return tmat

    def apply_matrix(self, transform_matrix):
        tmatrix = np.array(transform_matrix)
        self.transform_matrix = self.transform_matrix.dot(tmatrix)

    def reset_matrix(self):
        self.transform_matrix = np.identity(4)

    def print_matrix(self):
        print(self.transform_matrix)

    def shear_x(self, theta):
        shear_mat = np.identity(4)
        shear_mat[0, 1] = np.tan(theta)
        self.transform_matrix = self.transform_matrix.dot(shear_mat)
        return shear_mat

    def shear_y(self, theta):
        shear_mat = np.identity(4)
        shear_mat[1, 0] = np.tan(theta)
        self.transform_matrix = self.transform_matrix.dot(shear_mat)
        return shear_mat

    def reset_transforms(self):
        """Reset all transformations to their default state."""
        self.transform_matrix = np.identity(4)

    def translate(self, x, y, z=0):
        tmat = matrix.translation_matrix(x, y, z)
        self.transform_matrix = self.transform_matrix.dot(tmat)
        return tmat

    def rotate(self, theta, axis=np.array([0, 0, 1])):
        axis = np.array(axis[:])
        tmat = matrix.rotation_matrix(axis, theta)
        self.transform_matrix = self.transform_matrix.dot(tmat)
        return tmat



import unittest
import numpy as np

from p5.sketch.Vispy2DRenderer.shape import PShape
from p5.core.color import Color
from p5.pmath import PI

vertices = [(0, 0), (1, 0), (1, 1), (0, 1)]

quad = PShape(
    vertices=vertices,
    fill_color=Color(255),
    stroke_color=Color(0),
    stroke_weight=2,
    stroke_join=1,
    stroke_cap=1,
)

contour_shape = PShape(
    vertices=[(0, 0), (100, 0), (100, 100), (0, 100)],
    fill_color=Color(255),
    stroke_color=Color(0),
    stroke_weight=2,
    stroke_join=1,
    stroke_cap=1,
    contours=[[25, 25], [75, 25], [75, 75], [25, 75]],
)


class TestPShape(unittest.TestCase):
    def test_properties(self):
        self.assertTrue(
            np.array_equal(quad.vertices, np.array([(0, 0), (1, 0), (1, 1), (0, 1)]))
        )
        self.assertEqual(quad._fill, Color(255))
        self.assertEqual(quad._stroke, Color(0))
        self.assertEqual(quad._stroke_weight, 2)
        self.assertEqual(quad._stroke_cap, 1)
        self.assertEqual(quad._stroke_join, 1)

    def test_transforms(self):
        quad.translate(100, 100, 100)
        self.assertTrue(
            np.array_equal(
                quad._matrix,
                np.array(
                    [
                        [1.0, 0.0, 0.0, 100.0],
                        [0.0, 1.0, 0.0, 100.0],
                        [0.0, 0.0, 1.0, 100.0],
                        [0.0, 0.0, 0.0, 1.0],
                    ]
                ),
            )
        )
        quad.reset_matrix()

        quad.rotate_x(PI / 2)
        self.assertTrue(
            np.allclose(
                quad._matrix,
                np.array(
                    [
                        [1.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, -1.0, 0.0],
                        [0.0, 1.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 1.0],
                    ]
                ),
            )
        )
        quad.reset_matrix()

        quad.rotate_y(PI / 2)
        self.assertTrue(
            np.allclose(
                quad._matrix,
                np.array(
                    [
                        [0.0, 0.0, 1.0, 0.0],
                        [0.0, 1.0, 0.0, 0.0],
                        [-1.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 0.0, 1.0],
                    ]
                ),
            )
        )
        quad.reset_matrix()

        quad.rotate_z(PI / 2)
        self.assertTrue(
            np.allclose(
                quad._matrix,
                np.array(
                    [
                        [0.0, -1.0, 0.0, 0.0],
                        [1.0, 0.0, 0.0, 0.0],
                        [0.0, 0.0, 1.0, 0.0],
                        [0.0, 0.0, 0.0, 1.0],
                    ]
                ),
            )
        )
        quad.reset_matrix()

        quad.scale(2, 2)
        self.assertTrue(
            np.allclose(
                quad._matrix,
                np.array(
                    [
                        [2.0, 0.0, 0.0, 0.0],
                        [0.0, 2.0, 0.0, 0.0],
                        [0.0, 0.0, 1.0, 0.0],
                        [0.0, 0.0, 0.0, 1.0],
                    ]
                ),
            )
        )
        quad.reset_matrix()

        quad.shear_x(PI / 4)
        self.assertTrue(
            np.allclose(
                quad._matrix,
                np.array(
                    [
                        [1.0, 1.0, 0.0, 0.0],
                        [0.0, 1.0, 0.0, 0.0],
                        [0.0, 0.0, 1.0, 0.0],
                        [0.0, 0.0, 0.0, 1.0],
                    ]
                ),
            )
        )
        quad.reset_matrix()

        quad.shear_y(PI / 4)
        self.assertTrue(
            np.allclose(
                quad._matrix,
                np.array(
                    [
                        [1.0, 0.0, 0.0, 0.0],
                        [1.0, 1.0, 0.0, 0.0],
                        [0.0, 0.0, 1.0, 0.0],
                        [0.0, 0.0, 0.0, 1.0],
                    ]
                ),
            )
        )
        quad.reset_matrix()


if __name__ == "__main__":
    unittest.main()



# This file adds Processing API compatibility for p5py

from .structure import push_style, pop_style
from .primitives import rect_mode, ellipse_mode
from .transforms import (
    push_matrix,
    pop_matrix,
    reset_matrix,
    reset_transforms,
    shear_x,
    shear_y,
    rotate_x,
    rotate_y,
    rotate_z,
    apply_matrix,
    print_matrix,
)
from .color import color_mode
from .attribs import no_fill, no_tint, no_stroke, stroke_cap, stroke_join, stroke_weight
from .image import load_image, load_pixels, save_frame, image_mode
from .font import (
    create_font,
    load_font,
    text_font,
    text_size,
    text_width,
    text_descent,
    text_align,
    text_ascent,
    text_leading,
)
from .constants import TESS
from .vertex import (
    begin_shape,
    begin_contour,
    curve_vertex,
    bezier_vertex,
    end_contour,
    end_shape,
    quadratic_vertex,
)
from .material import normal_material, basic_material, blinn_phong_material
from .light import (
    ambient_light,
    directional_light,
    point_light,
    light_falloff,
    light_specular,
)


def push():
    """The push() function saves the current drawing style settings and transformations"""
    push_matrix()
    push_style()


def pop():
    pop_style()
    pop_matrix()


def pushStyle():
    """Save the current style settings and then restores them on exit.

    The 'style' information consists of all the parameters controlled
    by the following functions (the ones indicated by an asterisks '*'
    aren't available yet):

    - background
    - fill, noFill
    - stroke, noStroke
    - rectMode
    - ellipseMode
    - shapeMode
    - colorMode
    - tint
    - (*) strokeWeight
    - (*) strokeCap
    - (*) strokeJoin
    - (*) imageMode
    - (*) textAlign
    - (*) textFont
    - (*) textMode
    - (*) textSize
    - (*) textLeading
    -  emissive
    -  specular
    -  shininess
    -  ambient
    -  material

    """
    push_style()


def popStyle():
    """Restores previously pushed style settings"""
    pop_style()


def rectMode(mode="CORNER"):
    """Change the rect and square drawing mode for the p5.renderer.

    :param mode: The new mode for drawing rects. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'}. This defaults to
        'CORNER' so calling rect_mode without parameters will reset
        the sketch's rect mode.
    :type mode: str

    """
    rect_mode(mode)


def ellipseMode(mode="CENTER"):
    """Change the ellipse and circle drawing mode for the p5.renderer.

    :param mode: The new mode for drawing ellipses. Should be one of
        {'CORNER', 'CORNERS', 'CENTER', 'RADIUS'}. This defaults to
        'CENTER' so calling ellipse_mode without parameters will reset
        the sketch's ellipse mode.
    :type mode: str

    """
    ellipse_mode(mode)


def pushMatrix():
    """Pushes the current transformation matrix onto the matrix stack."""
    push_matrix()


def popMatrix():
    """Pops the current transformation matrix off the matrix stack."""
    pop_matrix()


def resetTransforms():
    """Reset all transformations to their default state."""
    reset_transforms()


def rotateX(theta):
    """Rotate the view along the x axis.

    :param theta: angle by which to rotate (in radians)
    :type theta: float

    :returns: The rotation matrix used to apply the transformation.
    :rtype: np.ndarray

    """
    return rotate_x(theta)


def rotateY(theta):
    """Rotate the view along the y axis.

    :param theta: angle by which to rotate (in radians)
    :type theta: float

    :returns: The rotation matrix used to apply the transformation.
    :rtype: np.ndarray

    """
    return rotate_y(theta)


def rotateZ(theta):
    """Rotate the view along the z axis.

    :param theta: angle by which to rotate (in radians)
    :type theta: float

    :returns: The rotation matrix used to apply the transformation.
    :rtype: np.ndarray

    """
    return rotate_z(theta)


def applyMatrix(transformMatrix):
    """Apply the given matrix to the sketch's transform matrix..

    :param transformMatrix: The new transform matrix.
    :type transformMatrix: np.ndarray (or a 4×4 list)
    """
    apply_matrix(transformMatrix)


def resetMatrix():
    """Reset the current transform matrix."""
    reset_matrix()


def printMatrix():
    """Print the transform matrix being used by the sketch."""
    print_matrix()


def shearX(theta):
    """Shear display along the x-axis.

    :param theta: angle to shear by (in radians)
    :type theta: float

    :returns: The shear matrix used to apply the tranformation.
    :rtype: np.ndarray

    """
    return shear_x(theta)


def shearY(theta):
    """Shear display along the y-axis.

    :param theta: angle to shear by (in radians)
    :type theta: float

    :returns: The shear matrix used to apply the transformation.
    :rtype: np.ndarray

    """
    return shear_y(theta)


def colorMode(mode, max1=255, max2=None, max3=None, maxAlpha=255):
    """Set the color mode of the renderer.

    :param mode: One of {'RGB', 'HSB'} corresponding to Red/Green/Blue
        or Hue/Saturation/Brightness
    :type mode: str

    :param max1: Maximum value for the first color channel (default:
        255)
    :type max1: int

    :param max2: Maximum value for the second color channel (default:
        max1)
    :type max2: int

    :param max3: Maximum value for the third color channel (default:
        max1)
    :type max3: int

    :param maxalpha: Maximum value for the alpha channel (default:
        255)
    :type maxalpha: int

    """
    color_mode(mode, max1, max2, max3, maxAlpha)


def noFill():
    """Disable filling geometry."""
    no_fill()


def strokeWeight(thickness):
    """Sets the width of the stroke used for lines, points, and the border around shapes. All widths are set in units of
     pixels.

    :param weight: thickness of stroke in pixels
    :type weight: int

    """
    stroke_weight(thickness)


def noStroke():
    """Disable drawing the stroke around shapes."""
    no_stroke()


def strokeCap(c):
    """Sets the style of line endings. The ends are SQUARE,
    PROJECT, and ROUND. The default cap is ROUND.

    :param c: either 'SQUARE', 'PROJECT' or 'ROUND'
    :type c: string

    """
    stroke_cap(c)


def strokeJoin(j):
    """Sets the style of the joints which connect line segments.
    These joints are either mitered, beveled, or rounded and
    specified with the corresponding parameters MITER, BEVEL,
    and ROUND. The default joint is MITER.

    :param weight: either 'MITER', 'BEVEL' or 'ROUND'
    :type j: string

    """
    stroke_join(j)


def noTint():
    """Disable tinting of images."""
    no_tint()


def imageMode(mode):
    """Modify the locaton from which the images are drawn.

    Modifies the location from which images are drawn by changing the
    way in which parameters given to :meth:`p5.image` are intepreted.

    The default mode is ``image_mode('corner')``, which interprets the
    second parameter of ``image()`` as the upper-left corner of the
    image. If an additional parameter is specified, it is used to set
    the image's width and height.

    ``image_mode('corners')`` interprets the first parameter of
    ``image()`` as the location of one corner, and the second
    parameter as the opposite corner.

    ``image_mode('center')`` interprets the first parameter of
    ``image()`` as the image's center point. If an additional
    parameter is specified, it is used to set the width and height of
    the image.

    :param mode: should be one of ``{'corner', 'center', 'corners'}``
    :type mode: str

    :raises ValueError: When the given image mode is not understood.
        Check for typoes.

    """
    image_mode(mode)


def loadImage(filename):
    """Load an image from the given filename.

    Loads an image into a variable of type PImage. Four types of
    images may be loaded.

    In most cases, load all images in setup() or outside the draw()
    call to preload them at the start of the program. Loading images
    inside draw() will reduce the speed of a program.

    :param filename: Filename (or path)of the given image. The
        file-extennsion is automatically inferred.
    :type filename: str

    :returns: An :class:`p5.PImage` instance with the given image data
    :rtype: :class:`p5.PImage`

    """
    return load_image(filename)


def loadPixels():
    """Load a snapshot of the display window into the ``pixels`` Image.

    This context manager loads data into the global ``pixels`` Image.
    Once the program execution leaves the context manager, all changes
    to the image are written to the main display.

    """
    load_pixels()


def saveFrame(filename=None):
    save_frame(filename)


def createFont(name, size=10):
    """Create the given font at the appropriate size.

    :param name: Filename of the font file (only pil, otf and ttf
        fonts are supported.)
    :type name: str

    :param size: Font size (only required when `name` refers to a
        truetype font; defaults to None)
    :type size: int | None

    """
    return create_font(name, size)


def loadFont(font_name):
    """Loads the given font into a font object"""
    return load_font(font_name)


def textFont(font, size=10):
    """Set current text font.

    :param font:
    :type font: PIL.ImageFont.ImageFont

    """
    text_font(font, size)


def textAlign(alignX, alignY=None):
    """Set the alignment of drawing text

    :param alignX: "RIGHT", "CENTER" or "LEFT".
    :type alignX: string

    :param alignY: "TOP", "CENTER" or "BOTTOM".
    :type alignY: string

    """
    text_align(alignX, alignY)


def textLeading(leading):
    """Sets the spacing between lines of text in units of pixels

    :param leading: the size in pixels for spacing between lines
    :type align_x: int

    """
    text_leading(leading)


def textSize(size):
    """Sets the current font size

    :param leading: the size of the letters in units of pixels
    :type align_x: int

    """
    text_size(size)


def textWidth(text):
    """Calculates and returns the width of any character or text string

    :param text_string: text
    :type text_string: str

    :returns: width of any character or text string
    :rtype: int

    """
    return text_width(text)


def textAscent():
    """Returns ascent of the current font at its current size

    :returns: ascent of the current font at its current size
    :rtype: float

    """
    return text_ascent()


def textDescent():
    """Returns descent of the current font at its current size

    :returns:  descent of the current font at its current size
    :rtype: float

    """
    return text_descent()


def beginShape(kind=TESS):
    """Begin shape drawing.  This is a helpful way of generating custom shapes quickly.

    :param kind: TESS, POINTS, LINES, TRIANGLES, TRIANGLE_FAN, TRIANGLE_STRIP, QUADS, or QUAD_STRIP; defaults to TESS
    :type kind: SType
    """
    begin_shape(kind)


def curveVertex(x, y, z=0):
    """
    Specifies vertex coordinates for curves. The first
    and last points in a series of curveVertex() lines
    will be used to guide the beginning and end of a the
    curve. A minimum of four points is required to draw a
    tiny curve between the second and third points. Adding
    a fifth point with curveVertex() will draw the curve
    between the second, third, and fourth points. The
    curveVertex() function is an implementation of
    Catmull-Rom splines.

    :param x: x-coordinate of the vertex
    :type x: float

    :param y: y-coordinate of the vertex
    :type y: float

    :param z: z-coordinate of the vertex
    :type z: float
    """
    curve_vertex(x, y, z)


def bezierVertex(x2, y2, x3, y3, x4, y4):
    """
    Specifies vertex coordinates for Bezier curves

    :param x2: x-coordinate of the first control point
    :type x2: float

    :param y2: y-coordinate of the first control point
    :type y2: float

    :param x3: x-coordinate of the second control point
    :type x3: float

    :param y3: y-coordinate of the second control point
    :type y3: float

    :param x4: x-coordinate of the anchor point
    :type x4: float

    :param y4: y-coordinate of the anchor point
    :type y4: float
    """
    bezier_vertex(x2, y2, x3, y3, x4, y4)


def quadraticVertex(cx, cy, x3, y3):
    """
    Specifies vertex coordinates for quadratic Bezier curves

    :param cx: x-coordinate of the control point
    :type cx: float

    :param cy: y-coordinate of the control point
    :type cy: float

    :param x3: x-coordinate of the anchor point
    :type x3: float

    :param y3: y-coordinate of the anchor point
    :type y3: float

    """
    quadratic_vertex(cx, cy, x3, y3)


def beginContour():
    """
    Use the beginContour() and endContour() functions
    to create negative shapes within shapes such as
    the center of the letter 'O'. beginContour() begins
    recording vertices for the shape and endContour() stops
    recording. The vertices that define a negative shape must
    "wind" in the opposite direction from the exterior shape.
    First draw vertices for the exterior clockwise order, then
    for internal shapes, draw vertices shape in counter-clockwise.

    """
    begin_contour()


def endContour():
    """Ends the current contour.

    For more info, see :any:`begin_contour`.
    """
    end_contour()


def endShape(mode=""):
    """
    The endShape() function is the companion to beginShape()
    and may only be called after beginShape(). When endshape()
    is called, all of image data defined since the previous call
    to beginShape() is rendered.

    :param mode: use CLOSE to close the shape
    :type mode: str

    """
    return end_shape(mode)


def normalMaterial():
    """The color is determined by the normal vector of a surface. Does not respond to lights.

    Useful for debugging.
    """
    normal_material()


def basicMaterial(r, g, b):
    """The default material. Always displays a solid color.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float
    """
    basic_material(r, g, b)


def blinnPhongMaterial():
    """Material based on the Blinn-Phong reflection model. This is the most "realistic" material in p5py.

    Blinn-Phong shading can be decomposed into three parts:
    ambient, diffuse, and specular.

    The ambient component is essentially a constant term that is alway present.
    We calculate it by summing all the ambient lights in a scene and multiplying it
    with the normalized ambient coefficent set by ambient.

    The diffuse component takes the normal vector of a surface into account and
    varies how much light is reflected depending on the angle that the surface
    makes with the incoming light.

    The specular component not only accounts for the direction of the light
    (like the diffuse component) but also the direction of the viewer. If the
    viewer is not on the path of the reflected light, the specular component
    falls off quickly, producing the glossy reflections we see on some materials.

    The color shown on the user's screen is the sum of all three components.
    """
    blinn_phong_material()


def ambientLight(r, g, b):
    """Adds an ambient light.

    Ambient light comes from all directions towards all directions.
    Ambient lights are almost always used in combination with other types of lights.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float
    """
    ambient_light(r, g, b)


def directionalLight(r, g, b, x, y, z):
    """Adds a directional light.
    Directional light comes from one direction: it is stronger when hitting a surface squarely,
    and weaker if it hits at a gentle angle.
    After hitting a surface, directional light scatters in all directions.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float

    :param x: x component of the direction vector
    :type x: float

    :param y: y component of the direction vector
    :type y: float

    :param z: z component of the direction vector
    :type z: float
    """
    directional_light(r, g, b, x, y, z)


def pointLight(r, g, b, x, y, z):
    """Adds a point light.
    Point light comes from one location and emits to all directions.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float

    :param x: x component of the location vector
    :type x: float

    :param y: y component of the location vector
    :type y: float

    :param z: z component of the location vector
    :type z: float
    """
    point_light(r, g, b, x, y, z)


def lightFalloff(constant, linear, quadratic):
    """Sets the falloff rates for point lights. Affects only the elements which are created after it in the code.

    d = distance from light position to vertex position

    falloff = 1 / (constant + d * linear + (d*d) * quadratic)

    If the coefficient is 0, then that term is ignored.
    The P3D renderer defaults to (0, 0, 0), i.e. no falloff.

    :param constant: coefficient for the constant term
    :type constant: float

    :param linear: coefficient for the linear term
    :type linear: float

    :param quadratic: coefficient for the quadratic term
    :type quadratic: float
    """
    light_falloff(constant, linear, quadratic)


def lightSpecular(r, g, b):
    """Sets the specular color for lights. Only visible with :any:`p5.blinn_phong_material`. Is set to (0 ,0, 0)
     by default.
    Affects only the elements which are created after it in the code.

    Specular refers to light which bounces off a surface in a preferred direction
    (rather than bouncing in all directions like a diffuse light) and is used for
    creating highlights. The specular quality of a light interacts with the specular
    material qualities set through the `specular()` and `shininess()` functions.

    :param r: red channel
    :type r: float

    :param g: green channel
    :type g: float

    :param b: blue channel
    :type b: float
    """
    light_specular(r, g, b)



# This file adds Processing API compatibility to p5py

from .local_storage import get_item, set_item, clear_storage, remove_item


def getItem(key):
    """
    Returns the value of an item that was stored int local storage using store_item(), None if not found

    :param key: key name that was used while storing data using store_item()
    :type key: string

    :returns: Value of stored item
    :rtype: object

    """
    return get_item(key)


def setItem(key, value):
    """
    * Stores a value in local storage under the key name.
    * Local storage persists in between the p5py sketch sessions.
    * The key can be the name of the variable but doesn't have to be.
    * Sensitive data such as passwords or personal information should not be stored in local storage.
    * Overwrites the key if already present

    :param key: key name that you wish to you use in local storage
    :type key: basestring

    :param value: Value to be stored, can be of any type
    :type value: object

    """
    set_item(key, value)


def removeItem(key):
    """
    Removes an item that was stored with store_item()

    :param key: key name to remove the associated value
    :type key: string

    """
    remove_item(key)


def clearStorage():
    """
    Clears all storage items set with store_item()
    """
    clear_storage()



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

__title__ = "p5"
__description__ = "Creative coding in Python"
__url__ = "https://p5py.github.io"
__version__ = "0.8.0"
__author__ = "Abhik Pal"
__author_email__ = "theabhikpal@gmail.com"
__license__ = " GNU GPLv3"
__copyright__ = "Copyright (C) 2017-2020 Abhik Pal"



"""
 * This module defines the p5 methods for working with local storage
"""

import pathlib
import pickle
import os

local_storage_path = os.path.join(pathlib.Path.home(), ".p5pyStorage")

if not os.path.isdir(local_storage_path):
    os.mkdir(local_storage_path)


def get_item(key):
    """
    Returns the value of an item that was stored int local storage using store_item(), None if not found

    :param key: key name that was used while storing data using store_item()
    :type key: string

    :returns: Value of stored item
    :rtype: object

    """
    item_file_path = os.path.join(local_storage_path, key)
    if os.path.isfile(item_file_path):
        with open(item_file_path, "rb") as f:
            return pickle.load(f)
    else:
        return None


def set_item(key, value):
    """
    * Stores a value in local storage under the key name.
    * Local storage persists in between the p5py sketch sessions.
    * The key can be the name of the variable but doesn't have to be.
    * Sensitive data such as passwords or personal information should not be stored in local storage.
    * Overwrites the key if already present

    :param key: key name that you wish to you use in local storage
    :type key: basestring

    :param value: Value to be stored, can be of any type
    :type value: object

    """
    item_file_path = os.path.join(local_storage_path, key)

    if os.path.isfile(item_file_path):
        os.remove(item_file_path)

    with open(item_file_path, "wb") as f:
        pickle.dump(value, f)


def remove_item(key):
    """
    Removes an item that was stored with store_item()

    :param key: key name to remove the associated value
    :type key: string

    """
    item_file_path = os.path.join(local_storage_path, key)

    if os.path.isfile(item_file_path):
        os.remove(item_file_path)


def clear_storage():
    """
    Clears all storage items set with store_item()
    """
    if os.path.isdir(local_storage_path):
        for filename in os.listdir(local_storage_path):
            file_path = os.path.join(local_storage_path, filename)
            os.remove(file_path)



from OpenGL.GLU import (
    gluNewTess,
    gluTessCallback,
    GLU_TESS_VERTEX,
    GLU_TESS_BEGIN,
    GLU_TESS_END,
    GLU_TESS_ERROR,
    gluErrorString,
    GLU_TESS_COMBINE,
)
from OpenGL.GL import GL_TRIANGLE_FAN, GL_TRIANGLE_STRIP, GL_TRIANGLES, GL_LINE_LOOP
import numpy as np


class Tessellator:
    def __init__(self):
        self.tess = gluNewTess()
        self.primitives = []  # [[gl_type, vertices, idx]...]
        self.vertices = []  # Vertices for the current shape
        self.type = None  # Type for the current shape

        gl_mode_map = {
            GL_TRIANGLE_FAN: "triangle_fan",
            GL_TRIANGLE_STRIP: "triangle_strip",
            GL_TRIANGLES: "triangles",
            GL_LINE_LOOP: "line_loop",
        }

        def end_shape_handler():
            curr_obj = [
                gl_mode_map[self.type],
                self.vertices,
                np.arange(len(self.vertices), dtype=np.uint32),
            ]
            self.primitives.append(curr_obj)

        def begin_shape_handler(x):
            self.type = x
            self.vertices = []

        def vertex_handler(v):
            self.vertices.append(v)

        def combine_handler(new_vert, _, __):
            return new_vert

        # Register Callbacks
        gluTessCallback(self.tess, GLU_TESS_VERTEX, vertex_handler)
        gluTessCallback(self.tess, GLU_TESS_END, end_shape_handler)
        gluTessCallback(
            self.tess, GLU_TESS_ERROR, lambda x: print("Error", gluErrorString(x))
        )
        gluTessCallback(self.tess, GLU_TESS_BEGIN, begin_shape_handler)
        gluTessCallback(self.tess, GLU_TESS_COMBINE, combine_handler)

    def get_result(self):
        """Returns the current primitive_list and clears it
        For the format of the draw queue, see the definition of self.primitive_list in `__init__`
        """
        res = self.primitives
        self.primitives = []
        return res



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from numpy.linalg import inv
from dataclasses import dataclass
from sys import stderr
import numpy as np
import math

import builtins

from vispy import gloo
from vispy.gloo import Texture2D, Program

from contextlib import contextmanager

from p5.core.constants import Z_EPSILON
from p5.core.geometry import Geometry
from ..Vispy2DRenderer.shape import PShape
from ..Vispy2DRenderer.openglrenderer import Style2D

from p5.pmath.matrix import translation_matrix
from p5.pmath import matrix
from ..Vispy2DRenderer.openglrenderer import (
    OpenGLRenderer,
    get_render_primitives,
    to_3x3,
    COLOR_WHITE,
)
from .shaders3d import src_default, src_fbuffer, src_normal, src_phong
from p5.core.material import BasicMaterial, NormalMaterial, BlinnPhongMaterial

from p5.core import p5


class GlslList:
    """List of objects to be used in glsl"""

    def __init__(self, max_size, obj_size, dtype):
        """Initialize GlslList
        max_size: The maximum size of the list
        obj_size: The length of an individual object
        dtype: The data type of this list
        """
        list_shape = (max_size, obj_size)
        self.data = np.zeros(list_shape, dtype=dtype)
        self.size = 0
        self.max_size = max_size

    def add(self, obj):
        if self.size == self.max_size:
            print(
                "Too many instances of {} are added. Max size {}.".format(
                    type(obj), self.max_size
                ),
                file=stderr,
            )
            return
        self.data[self.size] = obj
        self.size += 1

    def clear(self):
        self.data = np.zeros_like(self.data)
        self.size = 0


@dataclass
class Style3D(Style2D):
    ambient = np.array([0.2] * 3)
    diffuse = np.array([0.6] * 3)
    specular = np.array([0.8] * 3)
    shininess = 8
    material = BasicMaterial(COLOR_WHITE)


class Renderer3D(OpenGLRenderer):
    def __init__(self):
        super().__init__(src_fbuffer, src_default)
        self.style = Style3D()
        self.normal_prog = Program(src_normal.vert, src_normal.frag)
        self.phong_prog = Program(src_phong.vert, src_phong.frag)
        self.lookat_matrix = np.identity(4)

        # Camera position
        self.camera_pos = np.zeros(3)
        # Lights
        self.MAX_LIGHTS_PER_CATEGORY = 8
        self.ambient_light_color = GlslList(self.MAX_LIGHTS_PER_CATEGORY, 3, np.float32)
        self.directional_light_dir = GlslList(
            self.MAX_LIGHTS_PER_CATEGORY, 3, np.float32
        )
        self.directional_light_color = GlslList(
            self.MAX_LIGHTS_PER_CATEGORY, 3, np.float32
        )
        self.directional_light_specular = GlslList(
            self.MAX_LIGHTS_PER_CATEGORY, 3, np.float32
        )
        self.point_light_color = GlslList(self.MAX_LIGHTS_PER_CATEGORY, 3, np.float32)
        self.point_light_pos = GlslList(self.MAX_LIGHTS_PER_CATEGORY, 3, np.float32)
        self.point_light_specular = GlslList(
            self.MAX_LIGHTS_PER_CATEGORY, 3, np.float32
        )
        self.const_falloff = GlslList(self.MAX_LIGHTS_PER_CATEGORY, 1, np.float32)
        self.linear_falloff = GlslList(self.MAX_LIGHTS_PER_CATEGORY, 1, np.float32)
        self.quadratic_falloff = GlslList(self.MAX_LIGHTS_PER_CATEGORY, 1, np.float32)
        (
            self.curr_linear_falloff,
            self.curr_quadratic_falloff,
            self.curr_constant_falloff,
        ) = (0.0, 0.0, 0.0)
        self.light_specular = np.array([0.0] * 3)

    def reset_view(self):
        self.viewport = (
            0,
            0,
            int(builtins.width * builtins.pixel_x_density),
            int(builtins.height * builtins.pixel_y_density),
        )
        self.texture_viewport = (
            0,
            0,
            builtins.width,
            builtins.height,
        )

        gloo.set_viewport(*self.viewport)  # pylint: disable=no-member

        cz = (builtins.height / 2) / math.tan(math.radians(30))
        self.projection_matrix = matrix.perspective_matrix(
            math.radians(60), builtins.width / builtins.height, 0.1 * cz, 10 * cz
        )

        self.transform_matrix = np.identity(4)
        self._update_shader_transforms()

        self.fbuffer_tex_front = Texture2D((builtins.height, builtins.width, 3))
        self.fbuffer_tex_back = Texture2D((builtins.height, builtins.width, 3))
        self.fbuffer.depth_buffer = gloo.RenderBuffer((builtins.height, builtins.width))

        for buf in [self.fbuffer_tex_front, self.fbuffer_tex_back]:
            self.fbuffer.color_buffer = buf
            with self.fbuffer:
                self.clear()

    def clear(self, color=True, depth=True):
        """Clear the renderer background."""
        gloo.set_state(  # pylint: disable=no-member
            clear_color=self.style.background_color
        )
        gloo.clear(color=color, depth=depth)  # pylint: disable=no-member

    def clear_lights(self):
        self.ambient_light_color.clear()
        self.directional_light_color.clear()
        self.directional_light_dir.clear()
        self.directional_light_specular.clear()
        self.point_light_color.clear()
        self.point_light_pos.clear()
        self.point_light_specular.clear()
        self.const_falloff.clear()
        self.linear_falloff.clear()
        self.quadratic_falloff.clear()

    def _comm_toggles(self, state=True):
        gloo.set_state(blend=state)  # pylint: disable=no-member
        gloo.set_state(depth_test=state)  # pylint: disable=no-member

        if state:
            gloo.set_state(  # pylint: disable=no-member
                blend_func=("src_alpha", "one_minus_src_alpha")
            )
            gloo.set_state(depth_func="lequal")  # pylint: disable=no-member

    def _update_shader_transforms(self):
        # Default shader
        self.default_prog["projection"] = self.projection_matrix.T.flatten()
        self.default_prog["perspective_matrix"] = self.lookat_matrix.T.flatten()

        # Normal shader
        self.normal_prog["projection"] = self.projection_matrix.T.flatten()
        self.normal_prog["perspective"] = self.lookat_matrix.T.flatten()
        # This is a no-op, meaning that the normals stay in world space, which
        # matches the behavior in p5.js
        normal_transform = np.identity(3)
        # I think the transformation below takes the vertices to camera space, but
        # the results are funky, so it's probably incorrect? - ziyaointl, 2020/07/20
        # normal_transform = np.linalg.inv(self.projection_matrix[:3, :3] @ self.lookat_matrix[:3, :3])
        self.normal_prog["normal_transform"] = normal_transform.flatten()

        # Blinn-Phong Shader
        self.phong_prog["projection"] = self.projection_matrix.T.flatten()
        self.phong_prog["perspective"] = self.lookat_matrix.T.flatten()

    @contextmanager
    def draw_loop(self):
        """The main draw loop context manager."""

        self.transform_matrix = np.identity(4)
        self._update_shader_transforms()
        self.fbuffer.color_buffer = self.fbuffer_tex_back

        with self.fbuffer:
            gloo.set_viewport(*self.texture_viewport)  # pylint: disable=no-member
            self._comm_toggles()
            self.fbuffer_prog["texture"] = self.fbuffer_tex_front
            self.fbuffer_prog.draw("triangle_strip")
            self.clear(color=False, depth=True)
            self.clear_lights()

            yield

            self.flush_geometry()
            self.transform_matrix = np.identity(4)

        gloo.set_viewport(*self.viewport)  # pylint: disable=no-member
        self._comm_toggles(False)
        self.clear()
        self.fbuffer_prog["texture"] = self.fbuffer_tex_back
        self.fbuffer_prog.draw("triangle_strip")

        self.fbuffer_tex_front, self.fbuffer_tex_back = (
            self.fbuffer_tex_back,
            self.fbuffer_tex_front,
        )

    def _add_to_draw_queue_simple(self, stype, vertices, idx, color):
        """Adds shape of stype to draw queue"""
        self.draw_queue.append((stype, (vertices, idx, color, None, None)))

    def tnormals(self, shape):
        """Obtain a list of vertex normals in world coordinates"""
        if isinstance(shape.material, BasicMaterial):  # Basic shader doesn't need this
            return None
        return shape.vertex_normals @ np.linalg.inv(
            to_3x3(self.transform_matrix) @ to_3x3(shape.matrix)
        )

    def render(self, shape):
        if isinstance(shape, Geometry):
            n = len(shape.vertices)
            # Perform model transform
            # TODO: Investigate moving model transform from CPU to the GPU
            tverts = self._transform_vertices(
                np.hstack([shape.vertices, np.ones((n, 1))]),
                shape.matrix,
                self.transform_matrix,
            )
            tnormals = self.tnormals(shape)

            edges = shape.edges
            faces = shape.faces

            self.add_to_draw_queue(
                "poly",
                tverts,
                edges,
                faces,
                self.style.fill_color,
                self.style.stroke_color,
                tnormals,
                self.style.material,
            )

        elif isinstance(shape, PShape):
            fill = shape.fill.normalized if shape.fill else None
            stroke = shape.stroke.normalized if shape.stroke else None

            obj_list = get_render_primitives(shape)
            for obj in obj_list:
                stype, vertices, idx = obj
                # Transform vertices
                vertices = self._transform_vertices(
                    np.hstack([vertices, np.ones((len(vertices), 1))]),
                    shape._matrix,
                    self.transform_matrix,
                )
                # Add to draw queue
                self._add_to_draw_queue_simple(
                    stype, vertices, idx, stroke if stype == "lines" else fill
                )

    def add_to_draw_queue(
        self,
        stype,
        vertices,
        edges,
        faces,
        fill=None,
        stroke=None,
        normals=None,
        material=None,
    ):
        """Add the given vertex data to the draw queue.

        :param stype: type of shape to be added. Should be one of {'poly',
                'path', 'point'}
        :type stype: str

        :param vertices: (N, 3) array containing the vertices to be drawn.
        :type vertices: np.ndarray

        :param edges: (N, 2) array containing edges as tuples of indices
                into the vertex array. This can be None when not appropriate
                (eg. for points)
        :type edges: None | np.ndarray

        :param faces: (N, 3) array containing faces as tuples of indices
                into the vertex array. For 'point' and 'path' shapes, this can
                be None
        :type faces: np.ndarray

        :param fill: Fill color of the shape as a normalized RGBA tuple.
                When set to `None` the shape doesn't get a fill (default: None)
        :type fill: None | tuple

        :param stroke: Stroke color of the shape as a normalized RGBA
                tuple. When set to `None` the shape doesn't get stroke
                (default: None)
        :type stroke: None | tuple
        // TODO: Update documentation
        // TODO: Unite style-related attributes for both 2D and 3D under one material class
        """

        fill_shape = self.style.fill_enabled and not (fill is None)
        stroke_shape = self.style.stroke_enabled and not (stroke is None)

        if fill_shape and stype not in ["point", "path"]:
            idx = np.array(faces, dtype=np.uint32).ravel()
            self.draw_queue.append(
                ["triangles", (vertices, idx, fill, normals, material)]
            )

        if stroke_shape:
            if stype == "point":
                idx = np.arange(0, len(vertices), dtype=np.uint32)
                self.draw_queue.append(
                    ["points", (vertices, idx, stroke, normals, material)]
                )
            else:
                idx = np.array(edges, dtype=np.uint32).ravel()
                self.draw_queue.append(
                    ["lines", (vertices, idx, stroke, normals, material)]
                )

    def render_with_shaders(self, draw_type, draw_obj):
        vertices, idx, color, normals, material = draw_obj
        """Like render_default but is aware of shaders other than the basic one"""
        # 0. If material does not need normals nor extra info, strip them out
        # and use the method from superclass
        if (
            material is None
            or isinstance(material, BasicMaterial)
            or draw_type in ["points", "lines"]
        ):
            OpenGLRenderer.render_default(self, draw_type, [draw_obj[:3]])
            return

        # 1. Get the number of vertices
        num_vertices = len(vertices)

        # 2. Create empty buffers based on the number of vertices.
        #
        data = np.zeros(
            num_vertices, dtype=[("position", np.float32, 3), ("normal", np.float32, 3)]
        )

        # 3. Loop through all the shapes in the geometry queue adding
        # it's information to the buffer.
        #
        draw_indices = []
        data["position"][
            0:num_vertices,
        ] = np.array(vertices)
        draw_indices.append(idx)
        data["normal"][
            0:num_vertices,
        ] = np.array(normals)
        self.vertex_buffer.set_data(data)
        self.index_buffer.set_data(np.hstack(draw_indices))

        if isinstance(material, NormalMaterial):
            # 4. Bind the buffer to the shader.
            #
            self.normal_prog.bind(self.vertex_buffer)

            # 5. Draw the shape using the proper shape type and get rid of
            # the buffers.
            #
            self.normal_prog.draw(draw_type, indices=self.index_buffer)
        elif isinstance(material, BlinnPhongMaterial):
            self.phong_prog.bind(self.vertex_buffer)
            self.phong_prog["u_cam_pos"] = self.camera_pos
            # Material attributes
            self.phong_prog["u_ambient_color"] = material.ambient
            self.phong_prog["u_diffuse_color"] = material.diffuse
            self.phong_prog["u_specular_color"] = material.specular
            self.phong_prog["u_shininess"] = material.shininess
            # Directional lights
            self.phong_prog[
                "u_directional_light_count"
            ] = self.directional_light_color.size
            self.phong_prog["u_directional_light_dir"] = self.directional_light_dir.data
            self.phong_prog[
                "u_directional_light_color"
            ] = self.directional_light_color.data
            self.phong_prog[
                "u_directional_light_specular"
            ] = self.directional_light_specular.data
            # Ambient lights
            self.phong_prog["u_ambient_light_count"] = self.ambient_light_color.size
            self.phong_prog["u_ambient_light_color"] = self.ambient_light_color.data
            # Point lights
            self.phong_prog["u_point_light_count"] = self.point_light_color.size
            self.phong_prog["u_point_light_color"] = self.point_light_color.data
            self.phong_prog["u_point_light_pos"] = self.point_light_pos.data
            self.phong_prog["u_point_light_specular"] = self.point_light_specular.data
            # Point light falloffs
            self.phong_prog["u_const_falloff"] = self.const_falloff.data
            self.phong_prog["u_linear_falloff"] = self.linear_falloff.data
            self.phong_prog["u_quadratic_falloff"] = self.quadratic_falloff.data
            # Draw
            self.phong_prog.draw(draw_type, indices=self.index_buffer)
        else:
            raise NotImplementedError("Material not implemented")

    def flush_geometry(self):
        """Flush all the shape geometry from the draw queue to the GPU."""
        for index, shape in enumerate(self.draw_queue):
            current_shape, current_obj = (
                self.draw_queue[index][0],
                self.draw_queue[index][1],
            )
            # If current_shape is lines, bring it to the front by epsilon
            # to resolve z-fighting
            if current_shape == "lines":
                # line_transform is used whenever we render lines to break ties in depth
                # We transform the points to camera space, move them by
                # Z_EPSILON, and them move them back to world space
                line_transform = inv(self.lookat_matrix).dot(
                    translation_matrix(0, 0, Z_EPSILON).dot(self.lookat_matrix)
                )
                vertices = current_obj[0]
                current_obj = (
                    np.hstack([vertices, np.ones((vertices.shape[0], 1))]).dot(
                        line_transform.T
                    )[:, :3],
                    *current_obj[1:],
                )
            self.render_with_shaders(current_shape, current_obj)

        self.draw_queue = []

    def cleanup(self):
        super(Renderer3D, self).cleanup()
        self.normal_prog.delete()
        self.phong_prog.delete()

    def add_ambient_light(self, r, g, b):
        self.ambient_light_color.add(np.array((r, g, b)))

    def add_directional_light(self, r, g, b, x, y, z):
        self.directional_light_color.add(np.array((r, g, b)))
        self.directional_light_dir.add(np.array((x, y, z)))
        self.directional_light_specular.add(self.light_specular)

    def add_point_light(self, r, g, b, x, y, z):
        self.point_light_color.add(np.array((r, g, b)))
        self.point_light_pos.add(np.array((x, y, z)))
        self.point_light_specular.add(self.light_specular)
        self.const_falloff.add(self.curr_constant_falloff)
        self.linear_falloff.add(self.curr_linear_falloff)
        self.quadratic_falloff.add(self.curr_quadratic_falloff)

    def rotate_x(self, theta):
        self.rotate(theta, axis=np.array([1, 0, 0]))

    def rotate_y(self, theta):
        self.rotate(theta, axis=np.array([0, 1, 0]))

    def rotate_z(self, theta):
        self.rotate(theta, axis=np.array([0, 0, 1]))

    def camera(self, *args, **kwargs):
        def real_camera(
            position=(0, 0, p5.sketch.size[1] / math.tan(math.pi / 6)),
            target_position=(0, 0, 0),
            up_vector=(0, 1, 0),
        ):
            self.lookat_matrix = matrix.look_at(
                np.array(position), np.array(target_position), np.array(up_vector)
            )
            self.camera_pos = np.array(position)

        if len(args) == 9:  # If using non-tuple arguments
            kwargs["position"] = args[:3]
            kwargs["target_position"] = args[3:6]
            kwargs["up_vector"] = args[6:]
        elif len(args) <= 3:  # If using tuple arguments
            if len(args) >= 1:
                kwargs["position"] = args[0]
            if len(args) >= 2:
                kwargs["target_position"] = args[1]
            if len(args) >= 3:
                kwargs["up_vector"] = args[2]
        else:
            raise ValueError("Unexpected number of arguments passed to camera()")

        real_camera(**kwargs)

    def perspective(self, fovy, aspect, near, far):
        self.projection_matrix = matrix.perspective_matrix(fovy, aspect, near, far)

    def ortho(self, left, right, bottom, top, near, far):
        self.projection_matrix = np.array(
            [
                [2 / (right - left), 0, 0, -(right + left) / (right - left)],
                [0, 2 / (top - bottom), 0, -(top + bottom) / (top - bottom)],
                [0, 0, -2 / (far - near), -(far + near) / (far - near)],
                [0, 0, 0, 1],
            ]
        )

    def frustum(self):
        raise NotImplementedError

    def shape(self, vertices, contours, shape_type, *args):
        """Render a PShape"""
        self.render(PShape(vertices=vertices, contours=contours, shape_type=shape_type))



#
# Part of p5: A Python package based on Processing
# Copyright (C) 2017-2019 Abhik Pal
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
import textwrap

from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont
from PIL import ImageFilter
from PIL import ImageChops

from .image import image
from .image import PImage
from .structure import push_style

from . import p5

__all__ = [
    "create_font",
    "load_font",
    "text",
    "text_font",
    "text_align",
    "text_leading",
    "text_size",
    "text_width",
    "text_ascent",
    "text_descent",
]

_font_family = ImageFont.load_default()
_text_align_x = "LEFT"
_text_align_y = "TOP"
_text_leading = 0


def create_font(name, size=10):
    """Create the given font at the appropriate size.

    :param name: Filename of the font file (only pil, otf and ttf
        fonts are supported.)
    :type name: str

    :param size: Font size (only required when `name` refers to a
        truetype font; defaults to None)
    :type size: int | None

    """

    if name.endswith("ttf") or name.endswith("otf"):
        font = ImageFont.truetype(name, size)
    elif name.endswith("pil"):
        font = ImageFont.load(name)
    else:
        raise NotImplementedError("Font type not supported.")

    return font


def load_font(font_name):
    """Loads the given font into a font object"""
    return create_font(font_name)


def text(*args, wrap_at=None):
    """Draw the given text on the screen and save the image.

    :param text_string: text to display
    :type text_string: str

    :param x: x-coordinate of text
    :type x: float

    :param y: y-coordinate of text
    :type y: float

    :param z: z-coordinate of text
    :type z: float

    :param position: position of the text on the screen
    :type position: tuple

    :param wrap_at: specifies the text wrapping column (defaults to
        None)
    :type wrap_at: int

    :returns: actual text that was drawn to the image (when wrapping
        is not set, this is just the unmodified text_string)
    :rtype: str

    """
    if len(args) == 2:
        text_string, position = args
    elif len(args) == 3 or len(args) == 4:
        text_string, position = args[0], args[1:]
    else:
        raise ValueError("Unexpected number of arguments passed to text()")

    if len(text_string) == 0:
        return

    global _font_family, _text_leading

    multiline = False
    if not (wrap_at is None):
        text_string = textwrap.fill(text_string, wrap_at)
        size = _font_family.getsize_multiline(text_string)
        multiline = True
    elif "\n" in text_string:
        multiline = True
        size = list(_font_family.getsize_multiline(text_string))
        size[1] += _text_leading * text_string.count("\n")
    else:
        size = _font_family.getsize(text_string)

    is_stroke_valid = False  # True when stroke_weight != 0
    is_min_filter = False  # True when stroke_weight <0
    if p5.renderer.style.stroke_enabled:
        stroke_weight = p5.renderer.style.stroke_weight
        if stroke_weight < 0:
            stroke_weight = abs(stroke_weight)
            is_min_filter = True

        if stroke_weight > 0:
            if stroke_weight % 2 == 0:
                stroke_weight += 1
            is_stroke_valid = True

    if is_stroke_valid:
        new_size = list(map(lambda x: x + 2 * stroke_weight, size))
        is_stroke_valid = True
        text_xy = (stroke_weight, stroke_weight)
    else:
        new_size = size
        text_xy = (0, 0)

    canvas = Image.new("RGBA", new_size, color=(0, 0, 0, 0))
    canvas_draw = ImageDraw.Draw(canvas)

    if multiline:
        canvas_draw.multiline_text(
            text_xy, text_string, font=_font_family, spacing=_text_leading
        )
    else:
        canvas_draw.text(text_xy, text_string, font=_font_family)

    text_image = PImage(*new_size)
    text_image._img = canvas

    if is_stroke_valid:
        if is_min_filter:
            canvas_dilate = canvas.filter(ImageFilter.MinFilter(stroke_weight))
        else:
            canvas_dilate = canvas.filter(ImageFilter.MaxFilter(stroke_weight))
        canvas_stroke = ImageChops.difference(canvas, canvas_dilate)
        text_stroke_image = PImage(*new_size)
        text_stroke_image._img = canvas_stroke

    width, height = new_size
    position = list(position)
    if _text_align_x == "LEFT":
        position[0] += 0
    elif _text_align_x == "RIGHT":
        position[0] -= width
    elif _text_align_x == "CENTER":
        position[0] -= width / 2

    if _text_align_y == "TOP":
        position[1] += 0
    elif _text_align_y == "BOTTOM":
        position[1] -= height
    elif _text_align_y == "CENTER":
        position[1] -= height / 2

    with push_style():
        if p5.renderer.style.fill_enabled:
            p5.renderer.style.tint_enabled = True
            p5.renderer.style.tint_color = p5.renderer.style.fill_color
            image(text_image, position)
        if p5.renderer.style.stroke_enabled and is_stroke_valid:
            p5.renderer.style.tint_enabled = True
            p5.renderer.style.tint_color = p5.renderer.style.stroke_color
            image(text_stroke_image, position)

    return text_string


def text_font(font, size=10):
    """Set current text font.

    :param font:
    :type font: PIL.ImageFont.ImageFont

    """
    global _font_family
    _font_family = font


def text_align(align_x, align_y=None):
    """Set the alignment of drawing text

    :param align_x: "RIGHT", "CENTER" or "LEFT".
    :type align_x: string

    :param align_y: "TOP", "CENTER" or "BOTTOM".
    :type align_y: string

    """

    global _text_align_x, _text_align_y
    _text_align_x = align_x

    if align_y:
        _text_align_y = align_y


def text_leading(leading):
    """Sets the spacing between lines of text in units of pixels

    :param leading: the size in pixels for spacing between lines
    :type align_x: int

    """

    global _text_leading
    _text_leading = leading


def text_size(size):
    """Sets the current font size

    :param leading: the size of the letters in units of pixels
    :type align_x: int

    """

    global _font_family

    # reload the font with new size
    if hasattr(_font_family, "path"):
        if _font_family.path.endswith("ttf") or _font_family.path.endswith("otf"):
            _font_family = ImageFont.truetype(_font_family.path, size)
    else:
        raise ValueError("text_size is not supported for Bitmap Fonts")


def text_width(text):
    """Calculates and returns the width of any character or text string

    :param text_string: text
    :type text_string: str

    :returns: width of any character or text string
    :rtype: int

    """

    return _font_family.getsize(text)[0]


def text_ascent():
    """Returns ascent of the current font at its current size

    :returns: ascent of the current font at its current size
    :rtype: float

    """
    global _font_family
    ascent, descent = _font_family.getmetrics()
    return ascent


def text_descent():
    """Returns descent of the current font at its current size

    :returns:  descent of the current font at its current size
    :rtype: float

    """

    global _font_family
    ascent, descent = _font_family.getmetrics()
    return descent



# for loading tables in csv format
class Table:
    """Class to represent tabular data

    :param path: Path to the CSV file.
    :type path: string
    """

    def __init__(self, path, seperator):
        """
        Initializes Table object when given the path to a CSV file.
        """
        file = open(path, "r")
        lines = file.readlines()
        data = []
        for line in lines:
            fragment = line.split(seperator)
            x = fragment[len(fragment) - 1]
            x = x[:-1]
            fragment[len(fragment) - 1] = x
            data.append(fragment)
        self.data = data

    def get_row_count(self):
        """
        :returns: Number of rows in the read CSV.
        :rtype: int
        """
        return len(self.data)

    def getRowCount(self):
        """
        :returns: Number of rows in the read CSV.
        :rtype: int
        """
        return self.get_row_count()

    def get_column_count(self):
        """
        :returns: Number of columns in the read CSV.
        :rtype: int
        """
        return len(self.data[0])

    def getColumnCount(self):
        """
        :returns: Number of columns in the read CSV.
        :rtype: int
        """
        return self.get_column_count()

    def get_column(self, name):
        """
        :param name: Name of the required column
        :type name: string

        :returns: An entire column based on the column index.
        :rtype: list
        """
        count = 0
        for i in self.data[0]:
            if i == name:
                break
            count = count + 1
        column = []
        for item in self.data:
            column.append(item[count])
        return column

    def getColumn(self, name):
        """
        :param name: Name of the required column
        :type name: string

        :returns: An entire column based on the column index.
        :rtype: list
        """
        return self.get_column(name)

    def get_row(self, index):
        """
        Returns an entire row when given the row index.

        :param index: Name of the row
        :type index: string

        :returns: An entire row when given the row index.
        :rtype: list
        """
        for fragment in self.data:
            if fragment[0] == index:
                return fragment

    def getRow(self, index):
        """
        Returns an entire row when given the row index.

        :param index: Name of the row
        :type index: string

        :returns: An entire row when given the row index.
        :rtype: list
        """
        return self.get_row(index)

    def get_array(self):
        """
        :returns: the entire csv data as an multidimensional array.
        :rtype: list
        """
        return self.data

    def getArray(self):
        """
        :returns: the entire csv data as an multidimensional array.
        :rtype: list
        """
        return self.get_array()


def load_table(path, mode="csv"):
    """
    Calls Table class and returns a Table class object

    :param path: Path to file
    :type path: string

    :param mode: Type of File csv/ssv/tsv.
    :type mode: string

    :returns: A table object
    :rtype: object

    """
    if mode == "csv":
        seperator = ","
    if mode == "ssv":
        seperator = ";"
    if mode == "tsv":
        seperator = "\t"
    table = Table(path, seperator)
    return table
